"""orbit badge — generate quality badges for HuggingFace dataset cards."""

from __future__ import annotations

from orbit.utils.display import console


GRADE_COLORS = {
    "A": "brightgreen",
    "B": "green",
    "C": "yellow",
    "D": "orange",
    "F": "red",
}


def generate_badge_markdown(grade: str, score: int | None = None) -> str:
    """Generate shields.io badge markdown for an ORBIT grade."""
    color = GRADE_COLORS.get(grade, "lightgrey")
    label = f"Grade%20{grade}"
    if score is not None:
        label += f"%20({score}/100)"
    return f"![ORBIT Grade: {grade}](https://img.shields.io/badge/ORBIT-{label}-{color})"


def run_badge(dataset_repo: str, push: bool = False) -> None:
    """Generate and optionally push an ORBIT badge to a dataset card."""
    import sys

    from orbit.analyzer.dataset_loader import load_dataset_info, load_episode_data
    from orbit.analyzer.quality import compute_quality
    from orbit.analyzer.success_predictor import DatasetReadinessGrader

    console.print(f"\n[bold]Generating ORBIT badge for {dataset_repo}...[/bold]")

    # Load and analyze
    try:
        info = load_dataset_info(dataset_repo)
        episode_data = load_episode_data(dataset_repo, max_episodes=50)
    except Exception as e:
        console.print(f"[bold red]Error:[/bold red] Could not load dataset: {e}")
        sys.exit(1)

    # Extract action episodes
    action_episodes = []
    for ep_idx in sorted(episode_data["episode_index"].unique()):
        ep = episode_data[episode_data["episode_index"] == ep_idx]
        actions = ep["action"].tolist()
        if actions:
            import numpy as np
            action_episodes.append(np.array([a for a in actions if a is not None]))

    if not action_episodes:
        console.print("[bold red]Error:[/bold red] No action data found.")
        sys.exit(1)

    # Grade
    grader = DatasetReadinessGrader()
    result = grader.grade(info, action_episodes)
    badge_md = generate_badge_markdown(result.grade, result.score)

    console.print(f"\n[bold green]Badge generated:[/bold green]")
    console.print(f"  Grade: [bold]{result.grade}[/bold] (score: {result.score}/100)")
    console.print(f"\n  [dim]Markdown:[/dim]")
    console.print(f"  {badge_md}")

    if push:
        _push_badge(dataset_repo, badge_md)
    else:
        console.print(f"\n[dim]Add --push to update the dataset card on HuggingFace.[/dim]")


def _push_badge(dataset_repo: str, badge_md: str) -> None:
    """Push badge to HuggingFace dataset card."""
    try:
        from huggingface_hub import HfApi

        api = HfApi()

        # Read existing README
        try:
            readme_content = api.hf_hub_download(
                repo_id=dataset_repo, filename="README.md", repo_type="dataset"
            )
            with open(readme_content) as f:
                content = f.read()
        except Exception:
            content = f"---\n---\n# {dataset_repo}\n"

        # Remove existing ORBIT badge if present
        lines = content.split("\n")
        lines = [l for l in lines if "img.shields.io/badge/ORBIT" not in l]

        # Insert badge after YAML frontmatter
        insert_idx = 0
        in_frontmatter = False
        for i, line in enumerate(lines):
            if line.strip() == "---":
                if not in_frontmatter:
                    in_frontmatter = True
                else:
                    insert_idx = i + 1
                    break

        lines.insert(insert_idx, "")
        lines.insert(insert_idx + 1, badge_md)
        lines.insert(insert_idx + 2, "")

        new_content = "\n".join(lines)

        api.upload_file(
            path_or_fileobj=new_content.encode(),
            path_in_repo="README.md",
            repo_id=dataset_repo,
            repo_type="dataset",
            commit_message="Add ORBIT quality badge",
        )
        console.print("[bold green]\u2713[/bold green] Badge pushed to dataset card!")

    except Exception as e:
        console.print(f"[bold red]Error:[/bold red] Could not push badge: {e}")
        console.print("[dim]Make sure you have write access to the dataset.[/dim]")



"""Tests for the LLM demonstration generator."""

import numpy as np
import pytest

from orbit.generator.base import BaseSimulator, Trajectory
from orbit.generator.agent import DemoGeneratorAgent


class MockSimulator(BaseSimulator):
    """Mock simulator for testing."""

    def __init__(self, succeed_on_attempt=1):
        self._succeed_on = succeed_on_attempt
        self._attempt = 0
        self._state = {
            "qpos": np.zeros(6),
            "qvel": np.zeros(6),
            "site_xpos": np.zeros((1, 3)),
            "time": 0.0,
        }

    def get_state(self):
        return self._state.copy()

    def execute_action(self, action):
        self._state["qpos"] = self._state["qpos"] + action[:6] * 0.01
        self._state["time"] += 0.01
        return self.get_state()

    def reset(self):
        self._state = {
            "qpos": np.zeros(6),
            "qvel": np.zeros(6),
            "site_xpos": np.zeros((1, 3)),
            "time": 0.0,
        }
        self._attempt += 1
        return self.get_state()

    def get_api_description(self):
        return "Mock simulator with 6 DOF. action shape: (6,)"

    def check_success(self, task, state):
        return self._attempt >= self._succeed_on


class TestTrajectory:
    def test_empty_trajectory(self):
        traj = Trajectory()
        assert not traj.success
        assert traj.total_steps == 0
        states, actions = traj.to_arrays()
        assert states.size == 0
        assert actions.size == 0

    def test_trajectory_with_data(self):
        traj = Trajectory()
        traj.states = [np.array([1, 2, 3]), np.array([4, 5, 6])]
        traj.actions = [np.array([0.1, 0.2]), np.array([0.3, 0.4])]
        traj.success = True
        traj.total_steps = 2

        states, actions = traj.to_arrays()
        assert states.shape == (2, 3)
        assert actions.shape == (2, 2)


class TestMockSimulator:
    def test_reset_returns_state(self):
        sim = MockSimulator()
        state = sim.reset()
        assert "qpos" in state
        assert "qvel" in state

    def test_execute_action_updates_state(self):
        sim = MockSimulator()
        sim.reset()
        action = np.ones(6)
        new_state = sim.execute_action(action)
        assert not np.allclose(new_state["qpos"], 0)

    def test_api_description(self):
        sim = MockSimulator()
        desc = sim.get_api_description()
        assert "6 DOF" in desc


class TestDemoGeneratorAgent:
    def test_extract_code_from_python_block(self):
        agent = DemoGeneratorAgent()
        response = "Here's the code:\n```python\ndef run(sim):\n    return True\n```\nDone."
        code = agent._extract_code(response)
        assert "def run(sim):" in code
        assert "return True" in code

    def test_extract_code_from_generic_block(self):
        agent = DemoGeneratorAgent()
        response = "```\ndef run(sim):\n    return True\n```"
        code = agent._extract_code(response)
        assert "def run(sim):" in code

    def test_extract_code_bare(self):
        agent = DemoGeneratorAgent()
        response = "def run(sim):\n    return True"
        code = agent._extract_code(response)
        assert "def run(sim):" in code

    def test_execute_safely_success(self):
        agent = DemoGeneratorAgent()
        sim = MockSimulator(succeed_on_attempt=1)
        sim.reset()

        code = "def run(sim):\n    sim.execute_action(np.ones(6))\n    return True"
        traj, error = agent._execute_safely(code, sim, "test task")

        assert error is None
        assert traj is not None

    def test_execute_safely_error(self):
        agent = DemoGeneratorAgent()
        sim = MockSimulator()
        sim.reset()

        code = "def run(sim):\n    raise ValueError('test error')"
        traj, error = agent._execute_safely(code, sim, "test task")

        assert traj is None
        assert error is not None
        assert "ValueError" in error

    def test_execute_safely_syntax_error(self):
        agent = DemoGeneratorAgent()
        sim = MockSimulator()
        sim.reset()

        code = "def run(sim:\n    return True"  # syntax error
        traj, error = agent._execute_safely(code, sim, "test task")

        assert traj is None
        assert error is not None


class TestGenerationResult:
    def test_generation_result_import(self):
        from orbit.generator.generate import GenerationResult
        result = GenerationResult(
            task="test",
            simulator="mujoco",
            total_attempts=10,
            successful_episodes=5,
            failed_episodes=5,
        )
        assert result.successful_episodes == 5



"""Tests for the orbit cleaner module."""

import json
from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import pytest

from orbit.cleaner import CleaningPlan, build_cleaning_plan, print_dry_run


def _make_episode_data(n_episodes=10, ep_length=100, action_dim=6, add_bad=True):
    """Create synthetic episode data for testing."""
    rows = []
    for ep_idx in range(n_episodes):
        for frame_idx in range(ep_length):
            if add_bad and ep_idx == 0:
                # Episode 0: dead frames (no motion)
                action = [0.0] * action_dim
            elif add_bad and ep_idx == 1:
                # Episode 1: very short (will have only 10 frames)
                if frame_idx >= 10:
                    continue
                action = list(np.random.randn(action_dim) * 0.5)
            else:
                action = list(np.random.randn(action_dim) * 0.5)
            state = list(np.random.randn(action_dim) * 0.3)
            rows.append({
                "episode_index": ep_idx,
                "frame_index": frame_idx,
                "action": action,
                "state": state,
            })
    return pd.DataFrame(rows)


def _make_info(n_episodes=10, fps=30.0, action_dim=6):
    """Create a mock DatasetInfo."""
    info = MagicMock()
    info.repo_id = "test/dataset"
    info.num_episodes = n_episodes
    info.num_frames = n_episodes * 100
    info.fps = fps
    info.robot_type = "so100"
    info.camera_names = []
    info.video_keys = []
    info.shapes = {"action": [action_dim], "state": [action_dim]}
    info.duration_seconds = n_episodes * 100 / fps
    return info


class TestBuildCleaningPlan:
    """Test the cleaning plan builder."""

    def test_remove_episodes_classified_as_remove(self):
        """Episodes the ranker classifies as 'remove' should be in the removal list."""
        ep_data = _make_episode_data(n_episodes=10, add_bad=True)
        info = _make_info(n_episodes=10)
        plan = build_cleaning_plan("test/dataset", ep_data, info)
        # Episode 0 has all dead frames, should be flagged
        remove_indices = [ep["index"] for ep in plan.episodes_to_remove]
        assert 0 in remove_indices  # dead frames episode

    def test_aggressive_includes_review_episodes(self):
        """With --aggressive, 'review' episodes should also be removed."""
        ep_data = _make_episode_data(n_episodes=10, add_bad=True)
        info = _make_info(n_episodes=10)
        plan_standard = build_cleaning_plan("test/dataset", ep_data, info, aggressive=False)
        plan_aggressive = build_cleaning_plan("test/dataset", ep_data, info, aggressive=True)
        # Aggressive should remove at least as many as standard
        assert len(plan_aggressive.episodes_to_remove) >= len(plan_standard.episodes_to_remove)

    def test_dry_run_produces_plan_without_modification(self):
        """Building a plan should not modify anything — it's just analysis."""
        ep_data = _make_episode_data(n_episodes=5, add_bad=False)
        info = _make_info(n_episodes=5)
        plan = build_cleaning_plan("test/dataset", ep_data, info)
        assert isinstance(plan, CleaningPlan)
        assert plan.original_episodes == 5

    def test_cleaning_report_json_structure(self):
        """The cleaning plan should have the expected structure."""
        ep_data = _make_episode_data(n_episodes=10, add_bad=True)
        info = _make_info(n_episodes=10)
        plan = build_cleaning_plan("test/dataset", ep_data, info)
        assert plan.original_dataset == "test/dataset"
        assert plan.original_episodes == 10
        assert isinstance(plan.episodes_to_remove, list)
        assert isinstance(plan.dead_joints, list)
        assert plan.episodes_to_keep == plan.original_episodes - len(plan.episodes_to_remove)
        assert plan.cleaning_mode in ("standard", "aggressive")
        assert isinstance(plan.quality_before, float)
        assert isinstance(plan.quality_after, float)

    def test_episode_removal_has_reason(self):
        """Each removed episode should have a reason."""
        ep_data = _make_episode_data(n_episodes=10, add_bad=True)
        info = _make_info(n_episodes=10)
        plan = build_cleaning_plan("test/dataset", ep_data, info)
        for ep in plan.episodes_to_remove:
            assert "index" in ep
            assert "reason" in ep
            assert len(ep["reason"]) > 0

    def test_empty_episode_data(self):
        """Should handle empty data gracefully."""
        ep_data = pd.DataFrame()
        info = _make_info(n_episodes=0)
        plan = build_cleaning_plan("test/dataset", ep_data, info)
        assert plan.episodes_to_keep == 0
        assert len(plan.episodes_to_remove) == 0

    def test_clean_dataset_no_removals(self):
        """A clean dataset should have no episodes to remove."""
        # Create uniform, well-behaved data
        rows = []
        for ep_idx in range(20):
            for frame_idx in range(100):
                t = frame_idx / 100.0
                action = [np.sin(t + i * 0.5) * 0.5 for i in range(6)]
                state = [np.cos(t + i * 0.3) * 0.3 for i in range(6)]
                rows.append({
                    "episode_index": ep_idx,
                    "frame_index": frame_idx,
                    "action": action,
                    "state": state,
                })
        ep_data = pd.DataFrame(rows)
        info = _make_info(n_episodes=20)
        plan = build_cleaning_plan("test/dataset", ep_data, info)
        # With well-behaved data, removals should be minimal
        assert plan.episodes_to_keep >= 15  # at least 75% kept



"""Generate ready-to-use project templates (Makefile, scripts, GH workflows)."""

from __future__ import annotations

from pathlib import Path


def generate_makefile(dataset: str, policy: str, min_grade: str = "C") -> str:
    """Generate a Makefile for the training pipeline."""
    return f"""\
# Generated by: orbit init
# Makefile for robot policy training pipeline

DATASET = {dataset}
POLICY = {policy}
MIN_GRADE = {min_grade}

.PHONY: check train verify all clean coach

check:
\torbit gate $(DATASET) -p $(POLICY) --min-grade $(MIN_GRADE)

train: check
\tlerobot-train \\
\t  --dataset.repo_id=$(DATASET) \\
\t  --policy.type=$(POLICY) \\
\t  --batch_size=32 \\
\t  --steps=100000

verify: train
\torbit verify ./outputs/train/latest/ -d $(DATASET)

clean:
\torbit clean $(DATASET) -p $(POLICY) --aggressive

coach:
\torbit coach $(DATASET) --target 50 --policy $(POLICY)

all: check train verify
"""


def generate_train_script(dataset: str, policy: str, min_grade: str = "C") -> str:
    """Generate a training shell script."""
    return f"""\
#!/bin/bash
# Generated by: orbit init
# Training pipeline script for {dataset}

set -euo pipefail

DATASET="{dataset}"
POLICY="{policy}"
MIN_GRADE="{min_grade}"

echo "=== Step 1: Quality Gate ==="
orbit gate "$DATASET" -p "$POLICY" --min-grade "$MIN_GRADE" || {{
    echo "Quality gate failed. Run: orbit analyze $DATASET"
    exit 1
}}

echo "=== Step 2: Training ==="
lerobot-train \\
  --dataset.repo_id="$DATASET" \\
  --policy.type="$POLICY" \\
  --batch_size=32 \\
  --steps=100000

echo "=== Step 3: Verification ==="
orbit verify ./outputs/train/latest/ -d "$DATASET"

echo "=== Done ==="
"""


def generate_github_workflow(dataset: str, policy: str, min_grade: str = "C") -> str:
    """Generate a GitHub Actions workflow for training."""
    return f"""\
# Generated by: orbit init
name: Train Robot Policy
on:
  push:
    paths: ['data/**']
  workflow_dispatch:

jobs:
  quality-check:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v4
      - uses: actions/setup-python@v5
        with:
          python-version: '3.11'
      - run: pip install orbit-robotics[formats]
      - run: orbit gate {dataset} -p {policy} --min-grade {min_grade}

  train:
    needs: quality-check
    runs-on: [self-hosted, gpu]
    steps:
      - uses: actions/checkout@v4
      - run: pip install lerobot orbit-robotics
      - run: |
          lerobot-train \\
            --dataset.repo_id={dataset} \\
            --policy.type={policy} \\
            --batch_size=32 \\
            --steps=100000

  verify:
    needs: train
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v4
      - run: pip install orbit-robotics
      - run: orbit verify ./outputs/train/latest/ -d {dataset}
"""


def init_project(
    dataset: str,
    policy: str,
    output_dir: str = ".",
    min_grade: str = "C",
) -> list[str]:
    """Generate all project files.

    Returns:
        List of created file paths.
    """
    out = Path(output_dir)
    out.mkdir(parents=True, exist_ok=True)
    created = []

    # Makefile
    makefile = out / "Makefile"
    makefile.write_text(generate_makefile(dataset, policy, min_grade))
    created.append(str(makefile))

    # train.sh
    train_sh = out / "train.sh"
    train_sh.write_text(generate_train_script(dataset, policy, min_grade))
    train_sh.chmod(0o755)
    created.append(str(train_sh))

    # GitHub workflow
    gh_dir = out / ".github" / "workflows"
    gh_dir.mkdir(parents=True, exist_ok=True)
    workflow = gh_dir / "train.yml"
    workflow.write_text(generate_github_workflow(dataset, policy, min_grade))
    created.append(str(workflow))

    return created



"""Real-time data collection coach.

Watches a dataset directory as new episodes are added, analyzes each
new episode against the existing dataset, and gives real-time feedback.
"""

from __future__ import annotations

import logging
import time
from dataclasses import dataclass, field
from pathlib import Path

import numpy as np

logger = logging.getLogger(__name__)


@dataclass
class CoachFeedback:
    """Feedback on a newly collected episode."""

    episode_index: int
    verdict: str  # "good", "review", "redo"
    consistency_score: float
    issues: list[str] = field(default_factory=list)
    suggestions: list[str] = field(default_factory=list)
    dataset_progress: str = ""


@dataclass
class CoachState:
    """Running state of the collection coach."""

    target_episodes: int
    policy: str
    current_count: int = 0
    total_good: int = 0
    total_review: int = 0
    total_redo: int = 0
    baseline_metrics: dict = field(default_factory=dict)
    recent_jerk_trend: list[float] = field(default_factory=list)


def _compute_episode_metrics(episode: np.ndarray) -> dict:
    """Compute per-episode metrics for coaching comparison."""
    metrics = {}

    if episode.ndim != 2 or len(episode) < 3:
        return metrics

    # Path length
    diffs = np.diff(episode, axis=0)
    metrics["path_length"] = float(np.sum(np.sqrt(np.sum(diffs**2, axis=1))))

    # Jerk (smoothness indicator)
    if len(episode) >= 4:
        vel = np.diff(episode, axis=0)
        acc = np.diff(vel, axis=0)
        jerk = np.diff(acc, axis=0)
        metrics["jerk"] = float(np.mean(np.sqrt(np.sum(jerk**2, axis=1))))
    else:
        metrics["jerk"] = 0.0

    # Duration (frame count as proxy)
    metrics["length"] = len(episode)

    # Action range utilization
    action_range = np.ptp(episode, axis=0)
    metrics["range_util"] = float(np.mean(action_range))

    return metrics


def _compute_baseline(action_episodes: list[np.ndarray]) -> dict:
    """Compute baseline statistics from existing episodes."""
    all_metrics: dict[str, list[float]] = {}

    for ep in action_episodes:
        m = _compute_episode_metrics(ep)
        for k, v in m.items():
            all_metrics.setdefault(k, []).append(v)

    baseline = {}
    for k, values in all_metrics.items():
        arr = np.array(values)
        median = float(np.median(arr))
        mad = float(np.median(np.abs(arr - median)))
        baseline[k] = {"median": median, "mad": max(mad, 1e-8)}

    return baseline


def _score_episode(
    episode: np.ndarray,
    baseline: dict,
) -> CoachFeedback:
    """Score a new episode against the baseline."""
    metrics = _compute_episode_metrics(episode)
    issues = []
    z_scores = {}

    for k, v in metrics.items():
        if k in baseline:
            b = baseline[k]
            z = abs(v - b["median"]) / (b["mad"] * 1.4826)  # MAD-based z-score
            z_scores[k] = z

            if k == "jerk" and z > 2.0:
                ratio = v / max(b["median"], 1e-8)
                issues.append(f"Jerk {ratio:.1f}x higher than median — jerky motion detected")
            elif k == "path_length" and z > 2.0:
                ratio = v / max(b["median"], 1e-8)
                if ratio > 1:
                    issues.append(
                        f"Path length {(ratio - 1) * 100:.0f}% longer than typical — "
                        "consider more direct approach"
                    )
                else:
                    issues.append(
                        f"Path length {(1 - ratio) * 100:.0f}% shorter than typical — "
                        "may be incomplete"
                    )
            elif k == "length" and z > 2.5:
                ratio = v / max(b["median"], 1e-8)
                if ratio > 1.5:
                    issues.append(f"Episode {ratio:.1f}x longer than median — possible hesitation")
                elif ratio < 0.5:
                    issues.append(f"Episode {ratio:.1f}x shorter than median — may be incomplete")

    # Determine verdict
    max_z = max(z_scores.values()) if z_scores else 0
    if max_z > 3.5:
        verdict = "redo"
    elif max_z > 2.0 or len(issues) >= 2:
        verdict = "review"
    else:
        verdict = "good"

    # Consistency score (inverse of mean z-score)
    mean_z = np.mean(list(z_scores.values())) if z_scores else 0
    consistency = max(0.0, min(1.0, 1.0 - mean_z / 5.0))

    suggestions = []
    if "jerk" in z_scores and z_scores["jerk"] > 2.0:
        suggestions.append("Slow down near the end of the trajectory for smoother motion")
    if "path_length" in z_scores and z_scores["path_length"] > 2.0:
        suggestions.append("Try to follow the same path as your earlier demonstrations")

    return CoachFeedback(
        episode_index=-1,  # set by caller
        verdict=verdict,
        consistency_score=round(consistency, 2),
        issues=issues,
        suggestions=suggestions,
    )


def _detect_fatigue(state: CoachState) -> str | None:
    """Detect operator fatigue from jerk trends."""
    recent = state.recent_jerk_trend
    if len(recent) < 10:
        return None

    last_5 = recent[-5:]
    first_5 = recent[:5]
    mean_last = np.mean(last_5)
    mean_first = np.mean(first_5)

    if mean_last > mean_first * 1.5:
        return (
            "Your last 5 episodes have higher jerk than your first 5. "
            "Take a short break — fatigue affects demonstration quality."
        )
    return None


def watch_and_coach(
    dataset_path: str,
    target_episodes: int = 50,
    policy: str = "act",
    poll_interval: float = 2.0,
) -> None:
    """Watch a dataset directory and coach the operator in real-time.

    Args:
        dataset_path: Path to dataset directory.
        target_episodes: Target number of episodes to collect.
        policy: Target policy for quality context.
        poll_interval: Seconds between directory polls.
    """
    from orbit.utils.display import console

    ds_path = Path(dataset_path)
    if not ds_path.exists():
        console.print(f"[red]Dataset path not found: {ds_path}[/red]")
        return

    console.print(f"\n[bold]Collection Coach[/bold] — Target: {target_episodes} episodes for {policy.upper()}")
    console.print(f"[dim]Watching: {ds_path}[/dim]\n")

    # Load existing episodes
    action_episodes = _load_existing_episodes(ds_path)
    state = CoachState(
        target_episodes=target_episodes,
        policy=policy,
        current_count=len(action_episodes),
    )

    if action_episodes:
        state.baseline_metrics = _compute_baseline(action_episodes)
        console.print(f"[dim]Loaded {len(action_episodes)} existing episodes as baseline[/dim]")
    else:
        console.print("[dim]No existing episodes — will build baseline as you collect[/dim]")

    seen_files: set[str] = set()
    for f in ds_path.rglob("*"):
        if f.suffix in (".parquet", ".npz", ".npy"):
            seen_files.add(str(f))

    try:
        while state.current_count < target_episodes:
            new_files = []
            for f in ds_path.rglob("*"):
                fstr = str(f)
                if fstr not in seen_files and f.suffix in (".parquet", ".npz", ".npy"):
                    new_files.append(f)
                    seen_files.add(fstr)

            for new_file in sorted(new_files):
                ep = _load_single_episode(new_file)
                if ep is None:
                    continue

                state.current_count += 1
                action_episodes.append(ep)

                # Update baseline periodically
                if state.current_count % 10 == 0 and len(action_episodes) >= 5:
                    state.baseline_metrics = _compute_baseline(action_episodes)

                if state.baseline_metrics:
                    feedback = _score_episode(ep, state.baseline_metrics)
                    feedback.episode_index = state.current_count
                    feedback.dataset_progress = (
                        f"{state.current_count}/{target_episodes} episodes "
                        f"({state.current_count * 100 // target_episodes}%)"
                    )

                    _print_feedback(console, feedback, state)

                    # Track jerk for fatigue detection
                    metrics = _compute_episode_metrics(ep)
                    if "jerk" in metrics:
                        state.recent_jerk_trend.append(metrics["jerk"])

                    # Fatigue check
                    fatigue_msg = _detect_fatigue(state)
                    if fatigue_msg:
                        console.print(f"\n  [yellow]Tip: {fatigue_msg}[/yellow]\n")
                else:
                    console.print(
                        f"  Episode {state.current_count} recorded "
                        f"({len(ep)} frames) — building baseline..."
                    )
                    if len(action_episodes) >= 5:
                        state.baseline_metrics = _compute_baseline(action_episodes)
                        console.print("[dim]  Baseline computed from first 5 episodes[/dim]")

            time.sleep(poll_interval)

    except KeyboardInterrupt:
        console.print("\n[dim]Coach stopped.[/dim]")

    # Final summary
    console.print(f"\n[bold]Collection Complete[/bold]")
    console.print(f"  Episodes: {state.current_count}/{target_episodes}")
    console.print(f"  Good: {state.total_good}  Review: {state.total_review}  Redo: {state.total_redo}")
    if state.current_count >= target_episodes:
        console.print(f"\n  Next: orbit analyze {dataset_path} --policy {policy}")
    console.print()


def _print_feedback(console, feedback: CoachFeedback, state: CoachState) -> None:
    """Print episode feedback to console."""
    verdict_colors = {"good": "green", "review": "yellow", "redo": "red"}
    verdict_icons = {"good": "\u2713", "review": "\u26a0", "redo": "\u274c"}
    color = verdict_colors.get(feedback.verdict, "white")
    icon = verdict_icons.get(feedback.verdict, "?")

    console.print(
        f"\n  Episode {feedback.episode_index} recorded"
    )
    console.print(
        f"    [{color}]{icon} Consistency: {feedback.consistency_score:.2f}[/{color}] "
        f"— [{color}]{feedback.verdict.upper()}[/{color}]"
    )

    for issue in feedback.issues:
        console.print(f"    [yellow]\u26a0[/yellow] {issue}")

    for suggestion in feedback.suggestions:
        console.print(f"    [dim]\u2192 {suggestion}[/dim]")

    # Progress bar
    pct = state.current_count * 100 // state.target_episodes
    filled = pct // 5
    bar = "\u2588" * filled + "\u2591" * (20 - filled)
    console.print(f"\n  Progress: {bar} {feedback.dataset_progress}")

    if feedback.verdict == "good":
        state.total_good += 1
    elif feedback.verdict == "review":
        state.total_review += 1
    else:
        state.total_redo += 1


def _load_existing_episodes(ds_path: Path) -> list[np.ndarray]:
    """Load existing episodes from a dataset directory."""
    episodes = []

    # Try LeRobot format first
    try:
        from orbit.analyzer.dataset_loader import load_episode_data
        from orbit.analyzer.coverage import _extract_array

        ep_data = load_episode_data(str(ds_path), max_episodes=200)
        if ep_data is not None and not ep_data.empty:
            for _, ep_df in ep_data.groupby("episode_index"):
                if "action" in ep_df.columns:
                    arr = _extract_array(ep_df["action"])
                    if arr is not None and arr.size > 0:
                        episodes.append(arr)
            if episodes:
                return episodes
    except Exception:
        pass

    # Fallback: scan for numpy/parquet files
    for f in sorted(ds_path.rglob("*.npz")):
        ep = _load_single_episode(f)
        if ep is not None:
            episodes.append(ep)

    for f in sorted(ds_path.rglob("*.npy")):
        ep = _load_single_episode(f)
        if ep is not None:
            episodes.append(ep)

    return episodes


def _load_single_episode(path: Path) -> np.ndarray | None:
    """Load a single episode from a file."""
    try:
        if path.suffix == ".npz":
            data = np.load(path)
            for key in ("actions", "action", "qpos", "joint_positions"):
                if key in data:
                    arr = data[key]
                    if arr.ndim == 2 and arr.shape[0] >= 3:
                        return arr
        elif path.suffix == ".npy":
            arr = np.load(path)
            if arr.ndim == 2 and arr.shape[0] >= 3:
                return arr
        elif path.suffix == ".parquet":
            import pandas as pd

            df = pd.read_parquet(path)
            if "action" in df.columns:
                from orbit.analyzer.coverage import _extract_array

                arr = _extract_array(df["action"])
                if arr is not None and arr.ndim == 2 and arr.shape[0] >= 3:
                    return arr
    except Exception:
        pass
    return None



"""Tests for the AI-powered training debugger enhancements."""

import json
import math
from pathlib import Path
from unittest.mock import patch

import pytest

from orbit.debugger import (
    _analyze_dynamics,
    _llm_diagnosis,
    _parse_lerobot_logs,
    _rule_based_diagnosis,
)


class TestLLMDiagnosisWithDataset:
    """Test that dataset analysis results are passed to the LLM."""

    @patch("orbit.llm_utils.call_gemini")
    def test_dataset_context_included_in_prompt(self, mock_gemini):
        mock_gemini.return_value = "Root cause: test"

        dynamics = {
            "status": "plateau",
            "message": "Loss plateau",
            "total_steps": 50000,
            "loss_curve_summary": {"start": 0.8, "end": 0.31, "q25": 0.5, "q50": 0.4, "q75": 0.35},
            "findings_data": {"start_loss": 0.8, "end_loss": 0.31, "ratio": 0.39},
        }
        findings = [{"severity": "high", "finding": "plateau", "cause": "data"}]
        analysis_results = {
            "episode_count": 45,
            "consistency": {"overall": 0.62, "path_length_cv": 0.38},
            "signal_health": {"dead_dimensions": [6, 7], "clipping_joints": []},
            "action_divergence": 0.23,
            "outlier_episodes": [3, 17, 28],
        }

        result = _llm_diagnosis(
            dynamics, None, findings, "test/dataset", "fake-key",
            analysis_results=analysis_results,
        )

        assert result == "Root cause: test"
        # Check the prompt passed to Gemini includes dataset context
        call_args = mock_gemini.call_args
        prompt = call_args[0][0]
        assert "DATASET ANALYSIS" in prompt
        assert "45 episodes" in prompt
        assert "Dead joints: [6, 7]" in prompt
        assert "Action divergence: 0.23" in prompt
        assert "Outlier episodes: [3, 17, 28]" in prompt

    @patch("orbit.llm_utils.call_gemini")
    def test_no_dataset_context_when_none(self, mock_gemini):
        mock_gemini.return_value = "Root cause: test"

        dynamics = {
            "status": "converged_well",
            "message": "OK",
            "total_steps": 100000,
            "loss_curve_summary": {},
            "findings_data": {},
        }
        findings = []

        _llm_diagnosis(dynamics, None, findings, "test/dataset", "fake-key", analysis_results=None)

        prompt = mock_gemini.call_args[0][0]
        assert "DATASET ANALYSIS" not in prompt

    @patch("orbit.llm_utils.call_gemini")
    def test_readiness_grade_in_context(self, mock_gemini):
        mock_gemini.return_value = "diagnosis"

        dynamics = {
            "status": "plateau",
            "message": "Plateau",
            "total_steps": 50000,
            "loss_curve_summary": {},
            "findings_data": {},
        }
        analysis_results = {
            "readiness": {"grade": "D", "score": 35},
        }

        _llm_diagnosis(
            dynamics, None, [], "test/data", "key",
            analysis_results=analysis_results,
        )

        prompt = mock_gemini.call_args[0][0]
        assert "Readiness grade: D" in prompt


class TestDebugTrainingWithAI:
    """Test the debug_training function with --ai flag."""

    def test_debug_training_accepts_use_ai(self, tmp_path):
        """Verify the function signature accepts use_ai parameter."""
        from orbit.debugger import debug_training

        # Create a minimal log file
        log_file = tmp_path / "train.log"
        log_file.write_text("step:1000 loss:0.5 grdn:2.0\nstep:2000 loss:0.4 grdn:1.5\n")

        # Should not raise — just verify the parameter is accepted
        with patch("orbit.llm_utils.load_api_key", return_value=None):
            debug_training(str(tmp_path), use_ai=True)

    def test_debug_training_accepts_analysis_results(self, tmp_path):
        """Verify analysis_results parameter is accepted."""
        from orbit.debugger import debug_training

        log_file = tmp_path / "train.log"
        log_file.write_text("step:1000 loss:0.5 grdn:2.0\n")

        analysis = {"episode_count": 50, "consistency": {"overall": 0.8}}

        with patch("orbit.llm_utils.load_api_key", return_value=None):
            debug_training(str(tmp_path), analysis_results=analysis)



"""Raw directory loader for numpy/CSV/parquet files."""

from __future__ import annotations

import json
import logging
from pathlib import Path

import numpy as np
import pandas as pd

from orbit.analyzer.dataset_loader import DatasetInfo
from orbit.loaders.base import BaseLoader, normalize_dataframe, sample_episode_indices

logger = logging.getLogger(__name__)


class DirectoryLoader(BaseLoader):
    """Load datasets from directories containing numpy, CSV, or parquet files.

    Supports three directory structures:

    Structure A: One file per episode
        dataset/
        ├── episode_000.npz   # keys: "action" (T, A), "state" (T, D)
        ├── episode_001.npz
        └── ...

    Structure B: Flat arrays + episode boundaries
        dataset/
        ├── actions.npy       # (total_frames, A)
        ├── states.npy        # (total_frames, D)
        └── episodes.json     # [{"start": 0, "end": 50}, ...]

    Structure C: CSV/parquet with episode column
        dataset/
        └── data.parquet      # columns: episode_index, action_0..action_N, state_0..state_N
    """

    def can_load(self, source: str) -> bool:
        """True if source is a directory with numpy/CSV/parquet files (not LeRobot)."""
        p = Path(source).expanduser()
        if not p.is_dir():
            return False
        # Exclude LeRobot datasets
        if (p / "meta" / "info.json").exists():
            return False
        # Check for supported files
        has_npz = bool(list(p.glob("*.npz"))[:1])
        has_npy = bool(list(p.glob("*.npy"))[:1])
        has_csv = bool(list(p.glob("*.csv"))[:1])
        has_parquet = bool(list(p.glob("*.parquet"))[:1])
        return has_npz or has_npy or has_csv or has_parquet

    def load_info(self, source: str) -> DatasetInfo:
        """Infer dataset metadata from directory contents."""
        root = Path(source).expanduser().resolve()
        structure = self._detect_structure(root)

        if structure == "per_episode_npz":
            return self._info_from_npz(root)
        elif structure == "flat_arrays":
            return self._info_from_flat(root)
        elif structure == "tabular":
            return self._info_from_tabular(root)
        else:
            raise ValueError(
                f"Cannot determine dataset structure in {source}. "
                "Expected: episode_*.npz files, actions.npy + episodes.json, "
                "or data.parquet/data.csv"
            )

    def load_episodes(
        self,
        source: str,
        episode_indices: list[int] | None = None,
        max_episodes: int | None = None,
    ) -> pd.DataFrame:
        """Load episode data from directory."""
        if max_episodes is None:
            max_episodes = 100

        root = Path(source).expanduser().resolve()
        structure = self._detect_structure(root)

        if structure == "per_episode_npz":
            return self._load_npz(root, episode_indices, max_episodes)
        elif structure == "flat_arrays":
            return self._load_flat(root, episode_indices, max_episodes)
        elif structure == "tabular":
            return self._load_tabular(root, episode_indices, max_episodes)
        else:
            raise ValueError(f"Cannot determine dataset structure in {source}")

    def _detect_structure(self, root: Path) -> str | None:
        """Detect which directory structure is present."""
        # Structure A: Per-episode npz files
        npz_files = sorted(root.glob("episode_*.npz")) + sorted(root.glob("ep_*.npz"))
        if npz_files:
            return "per_episode_npz"

        # Structure B: Flat arrays
        if (root / "actions.npy").exists():
            return "flat_arrays"

        # Structure C: Tabular (parquet or CSV)
        parquet_files = list(root.glob("*.parquet"))
        csv_files = list(root.glob("*.csv"))
        if parquet_files or csv_files:
            return "tabular"

        # Also check for any npz files (not episode_* named)
        any_npz = list(root.glob("*.npz"))
        if any_npz:
            return "per_episode_npz"

        return None

    # --- Structure A: Per-episode NPZ ---

    def _info_from_npz(self, root: Path) -> DatasetInfo:
        npz_files = sorted(root.glob("*.npz"))

        num_episodes = len(npz_files)
        num_frames = 0
        shapes: dict[str, list[int]] = {}

        if npz_files:
            sample = np.load(npz_files[0])
            if "action" in sample:
                shapes["action"] = list(sample["action"].shape[1:])
                num_frames = sum(
                    np.load(f)["action"].shape[0] for f in npz_files[:10]
                )
                # Estimate total from sample
                if len(npz_files) > 10:
                    avg_len = num_frames / 10
                    num_frames = int(avg_len * num_episodes)
            if "state" in sample:
                shapes["state"] = list(sample["state"].shape[1:])

        return DatasetInfo(
            repo_id=str(root),
            num_episodes=num_episodes,
            num_frames=num_frames,
            fps=30.0,  # Default; user can override
            robot_type="unknown",
            camera_names=[],
            video_keys=[],
            shapes=shapes,
        )

    def _load_npz(
        self, root: Path, episode_indices: list[int] | None, max_episodes: int
    ) -> pd.DataFrame:
        npz_files = sorted(root.glob("*.npz"))

        if episode_indices is not None:
            npz_files = [npz_files[i] for i in episode_indices if i < len(npz_files)]
        elif len(npz_files) > max_episodes:
            sampled = sample_episode_indices(len(npz_files), max_episodes)
            npz_files = [npz_files[i] for i in sampled]

        rows: list[dict] = []
        for ep_idx, f in enumerate(npz_files):
            data = np.load(f)
            actions = data.get("action", data.get("actions"))
            states = data.get("state", data.get("states"))

            n_frames = actions.shape[0] if actions is not None else (
                states.shape[0] if states is not None else 0
            )

            for frame_idx in range(n_frames):
                rows.append(
                    {
                        "episode_index": ep_idx,
                        "frame_index": frame_idx,
                        "action": (
                            actions[frame_idx].tolist() if actions is not None else None
                        ),
                        "state": (
                            states[frame_idx].tolist() if states is not None else None
                        ),
                        "timestamp": frame_idx / 30.0,
                    }
                )

        return normalize_dataframe(pd.DataFrame(rows) if rows else pd.DataFrame())

    # --- Structure B: Flat arrays ---

    def _info_from_flat(self, root: Path) -> DatasetInfo:
        actions = np.load(root / "actions.npy")
        states_path = root / "states.npy"
        states = np.load(states_path) if states_path.exists() else None

        episodes_file = root / "episodes.json"
        if episodes_file.exists():
            with open(episodes_file) as f:
                ep_boundaries = json.load(f)
            num_episodes = len(ep_boundaries)
        else:
            num_episodes = 1

        shapes: dict[str, list[int]] = {"action": list(actions.shape[1:])}
        if states is not None:
            shapes["state"] = list(states.shape[1:])

        return DatasetInfo(
            repo_id=str(root),
            num_episodes=num_episodes,
            num_frames=actions.shape[0],
            fps=30.0,
            robot_type="unknown",
            camera_names=[],
            video_keys=[],
            shapes=shapes,
        )

    def _load_flat(
        self, root: Path, episode_indices: list[int] | None, max_episodes: int
    ) -> pd.DataFrame:
        actions = np.load(root / "actions.npy")
        states_path = root / "states.npy"
        states = np.load(states_path) if states_path.exists() else None

        episodes_file = root / "episodes.json"
        if episodes_file.exists():
            with open(episodes_file) as f:
                ep_boundaries = json.load(f)
        else:
            ep_boundaries = [{"start": 0, "end": actions.shape[0]}]

        if episode_indices is not None:
            target_eps = [
                (i, ep_boundaries[i])
                for i in episode_indices
                if i < len(ep_boundaries)
            ]
        elif len(ep_boundaries) > max_episodes:
            sampled = sample_episode_indices(len(ep_boundaries), max_episodes)
            target_eps = [(i, ep_boundaries[i]) for i in sampled]
        else:
            target_eps = list(enumerate(ep_boundaries))

        rows: list[dict] = []
        for ep_idx, bounds in target_eps:
            start = bounds["start"]
            end = bounds["end"]
            for frame_idx, global_idx in enumerate(range(start, end)):
                rows.append(
                    {
                        "episode_index": ep_idx,
                        "frame_index": frame_idx,
                        "action": actions[global_idx].tolist(),
                        "state": (
                            states[global_idx].tolist() if states is not None else None
                        ),
                        "timestamp": frame_idx / 30.0,
                    }
                )

        return normalize_dataframe(pd.DataFrame(rows) if rows else pd.DataFrame())

    # --- Structure C: Tabular ---

    def _info_from_tabular(self, root: Path) -> DatasetInfo:
        df = self._read_tabular(root)

        num_episodes = (
            df["episode_index"].nunique() if "episode_index" in df.columns else 1
        )
        num_frames = len(df)

        # Detect action/state columns
        action_cols = sorted([c for c in df.columns if c.startswith("action")])
        state_cols = sorted([c for c in df.columns if c.startswith("state")])
        shapes: dict[str, list[int]] = {}
        if action_cols:
            shapes["action"] = [len(action_cols)]
        if state_cols:
            shapes["state"] = [len(state_cols)]

        return DatasetInfo(
            repo_id=str(root),
            num_episodes=num_episodes,
            num_frames=num_frames,
            fps=30.0,
            robot_type="unknown",
            camera_names=[],
            video_keys=[],
            shapes=shapes,
        )

    def _load_tabular(
        self, root: Path, episode_indices: list[int] | None, max_episodes: int
    ) -> pd.DataFrame:
        df = self._read_tabular(root)

        # Detect action/state columns
        action_cols = sorted([c for c in df.columns if c.startswith("action_")])
        state_cols = sorted([c for c in df.columns if c.startswith("state_")])

        # If the df already has 'action' and 'state' as single columns, use directly
        has_action_col = "action" in df.columns and not action_cols
        has_state_col = "state" in df.columns and not state_cols

        # Ensure episode_index exists
        if "episode_index" not in df.columns:
            df["episode_index"] = 0

        # Filter episodes
        all_episodes = sorted(df["episode_index"].unique())
        if episode_indices is not None:
            df = df[df["episode_index"].isin(episode_indices)]
        elif len(all_episodes) > max_episodes:
            sampled = sample_episode_indices(len(all_episodes), max_episodes)
            target = [all_episodes[i] for i in sampled if i < len(all_episodes)]
            df = df[df["episode_index"].isin(target)]

        # Combine action/state columns into lists
        rows: list[dict] = []
        for _, row in df.iterrows():
            action = None
            if has_action_col:
                action = row["action"]
            elif action_cols:
                action = [float(row[c]) for c in action_cols]

            state = None
            if has_state_col:
                state = row["state"]
            elif state_cols:
                state = [float(row[c]) for c in state_cols]

            rows.append(
                {
                    "episode_index": int(row["episode_index"]),
                    "frame_index": int(row.get("frame_index", 0)),
                    "action": action,
                    "state": state,
                    "timestamp": float(row.get("timestamp", 0.0)),
                }
            )

        return normalize_dataframe(pd.DataFrame(rows) if rows else pd.DataFrame())

    def _read_tabular(self, root: Path) -> pd.DataFrame:
        """Read the first parquet or CSV file found."""
        parquet_files = sorted(root.glob("*.parquet"))
        if parquet_files:
            return pd.read_parquet(parquet_files[0])

        csv_files = sorted(root.glob("*.csv"))
        if csv_files:
            return pd.read_csv(csv_files[0])

        raise ValueError(f"No parquet or CSV files found in {root}")



"""Tests for the plugin system."""

import tempfile
from pathlib import Path

import numpy as np

from orbit.plugins import BaseAnalyzer, AnalyzerResult, discover_plugins, run_plugins


def test_discover_empty_dir():
    with tempfile.TemporaryDirectory() as tmpdir:
        plugins = discover_plugins(Path(tmpdir))
        assert plugins == []


def test_discover_nonexistent_dir():
    plugins = discover_plugins(Path("/nonexistent/path/12345"))
    assert plugins == []


def test_discover_and_run_plugin():
    with tempfile.TemporaryDirectory() as tmpdir:
        plugin_file = Path(tmpdir) / "test_analyzer.py"
        plugin_file.write_text("""
from orbit.plugins import BaseAnalyzer, AnalyzerResult

class TestAnalyzer(BaseAnalyzer):
    name = "test_metric"
    description = "A test analyzer"

    def analyze(self, action_episodes, info=None, **kwargs):
        return AnalyzerResult(
            name=self.name,
            score=0.9,
            status="ok",
            summary="All good",
            details={"count": len(action_episodes)},
        )
""")
        plugins = discover_plugins(Path(tmpdir))
        assert len(plugins) == 1
        assert plugins[0].name == "test_metric"

        # Run plugins
        episodes = [np.random.randn(10, 6) for _ in range(5)]
        results = run_plugins(episodes, plugins_dir=Path(tmpdir))
        assert len(results) == 1
        assert results[0].score == 0.9
        assert results[0].details["count"] == 5


def test_broken_plugin_handled_gracefully():
    with tempfile.TemporaryDirectory() as tmpdir:
        plugin_file = Path(tmpdir) / "broken.py"
        plugin_file.write_text("""
from orbit.plugins import BaseAnalyzer, AnalyzerResult

class BrokenAnalyzer(BaseAnalyzer):
    name = "broken"

    def analyze(self, action_episodes, info=None, **kwargs):
        raise ValueError("intentional error")
""")
        episodes = [np.random.randn(10, 6)]
        results = run_plugins(episodes, plugins_dir=Path(tmpdir))
        assert len(results) == 1
        assert results[0].status == "bad"
        assert "intentional error" in results[0].summary



"""Tests for project initialization templates."""

from pathlib import Path

import pytest

from orbit.training.templates import (
    generate_github_workflow,
    generate_makefile,
    generate_train_script,
    init_project,
)


class TestMakefile:
    def test_contains_targets(self):
        makefile = generate_makefile("test/dataset", "act")
        assert "check:" in makefile
        assert "train:" in makefile
        assert "verify:" in makefile
        assert "all:" in makefile
        assert "orbit gate" in makefile
        assert "lerobot-train" in makefile

    def test_uses_provided_params(self):
        makefile = generate_makefile("my/data", "diffusion_policy", "B")
        assert "DATASET = my/data" in makefile
        assert "POLICY = diffusion_policy" in makefile
        assert "MIN_GRADE = B" in makefile


class TestTrainScript:
    def test_contains_pipeline_steps(self):
        script = generate_train_script("test/dataset", "act")
        assert "orbit gate" in script
        assert "lerobot-train" in script
        assert "orbit verify" in script
        assert "#!/bin/bash" in script

    def test_exit_on_gate_failure(self):
        script = generate_train_script("test/dataset", "act")
        assert "exit 1" in script


class TestGitHubWorkflow:
    def test_contains_jobs(self):
        workflow = generate_github_workflow("test/dataset", "act")
        assert "quality-check:" in workflow
        assert "train:" in workflow
        assert "verify:" in workflow
        assert "orbit gate" in workflow
        assert "needs: quality-check" in workflow

    def test_uses_correct_dataset(self):
        workflow = generate_github_workflow("my/data", "diffusion_policy")
        assert "my/data" in workflow
        assert "diffusion_policy" in workflow


class TestInitProject:
    def test_creates_all_files(self, tmp_path):
        created = init_project("test/data", "act", str(tmp_path))
        assert len(created) == 3

        makefile = tmp_path / "Makefile"
        assert makefile.exists()
        assert "orbit gate" in makefile.read_text()

        train_sh = tmp_path / "train.sh"
        assert train_sh.exists()
        # Check executable permission
        import stat
        assert train_sh.stat().st_mode & stat.S_IXUSR

        workflow = tmp_path / ".github" / "workflows" / "train.yml"
        assert workflow.exists()
        assert "quality-check" in workflow.read_text()

    def test_creates_output_dir(self, tmp_path):
        out = tmp_path / "new_project"
        init_project("test/data", "act", str(out))
        assert out.exists()



"""Configuration file support for ORBIT (~/.orbit/config.yaml)."""

from __future__ import annotations

import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any


CONFIG_DIR = Path.home() / ".orbit"
CONFIG_FILE = CONFIG_DIR / "config.yaml"


@dataclass
class OrbitConfig:
    """User configuration loaded from ~/.orbit/config.yaml."""

    default_policy: str = "auto"
    default_min_grade: str = "C"
    gpu_memory: int | None = None
    google_api_key: str | None = None
    default_framework: str = "lerobot"
    default_format: str | None = None
    default_device: str | None = None
    coach: dict = field(default_factory=lambda: {"target_episodes": 50, "poll_interval": 2.0})
    plugins_dir: str | None = None

    @classmethod
    def load(cls) -> "OrbitConfig":
        """Load config from ~/.orbit/config.yaml, falling back to defaults."""
        if not CONFIG_FILE.exists():
            return cls()

        try:
            import yaml
        except ImportError:
            # yaml not installed — silently use defaults
            return cls()

        try:
            with open(CONFIG_FILE) as f:
                data = yaml.safe_load(f) or {}
        except Exception:
            return cls()

        return cls._from_dict(data)

    @classmethod
    def _from_dict(cls, data: dict[str, Any]) -> "OrbitConfig":
        """Build config from a parsed YAML dict, expanding env vars."""
        cfg = cls()
        for key in [
            "default_policy", "default_min_grade", "gpu_memory",
            "google_api_key", "default_framework", "default_format",
            "default_device", "plugins_dir",
        ]:
            if key in data:
                val = data[key]
                if isinstance(val, str) and val.startswith("${") and val.endswith("}"):
                    val = os.environ.get(val[2:-1], None)
                setattr(cfg, key, val)
        if "coach" in data and isinstance(data["coach"], dict):
            cfg.coach = {**cfg.coach, **data["coach"]}
        return cfg

    def save(self) -> None:
        """Save current config to ~/.orbit/config.yaml."""
        try:
            import yaml
        except ImportError:
            raise RuntimeError("PyYAML required to save config: pip install pyyaml")

        CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        data = {
            "default_policy": self.default_policy,
            "default_min_grade": self.default_min_grade,
            "default_framework": self.default_framework,
        }
        if self.gpu_memory is not None:
            data["gpu_memory"] = self.gpu_memory
        if self.google_api_key is not None:
            data["google_api_key"] = self.google_api_key
        if self.default_format is not None:
            data["default_format"] = self.default_format
        if self.default_device is not None:
            data["default_device"] = self.default_device
        if self.plugins_dir is not None:
            data["plugins_dir"] = self.plugins_dir
        if self.coach != {"target_episodes": 50, "poll_interval": 2.0}:
            data["coach"] = self.coach

        with open(CONFIG_FILE, "w") as f:
            yaml.dump(data, f, default_flow_style=False, sort_keys=False)


_VALID_VALUES: dict[str, list[str] | None] = {
    "default_policy": ["auto", "act", "diffusion_policy", "smolvla", "bc", "dp3"],
    "default_min_grade": ["A", "B", "C", "D"],
    "default_framework": ["lerobot"],
    "default_format": [None, "auto", "lerobot", "lerobot-local", "rlds", "hdf5", "rosbag", "directory"],
    "default_device": [None, "cpu", "cuda", "mps"],
}


def validate_config_value(key: str, value: Any) -> str | None:
    """Validate a config value before setting. Returns error message or None."""
    allowed = _VALID_VALUES.get(key)
    if allowed is not None and value not in allowed:
        options = [str(v) for v in allowed if v is not None]
        return f"Invalid value '{value}' for {key}. Allowed: {', '.join(options)}"

    if key == "gpu_memory" and value is not None:
        if not isinstance(value, (int, float)) or value <= 0:
            return f"gpu_memory must be a positive number, got {value}"

    return None


def get_config() -> OrbitConfig:
    """Get the global config (loads once, caches)."""
    if not hasattr(get_config, "_cached"):
        get_config._cached = OrbitConfig.load()
    return get_config._cached



"""Plugin system for custom ORBIT analyzers."""

from __future__ import annotations

import importlib.util
import logging
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import numpy as np

logger = logging.getLogger(__name__)

DEFAULT_PLUGINS_DIR = Path.home() / ".orbit" / "plugins"


class BaseAnalyzer(ABC):
    """Base class for custom ORBIT analyzer plugins.

    Subclass this and place in ~/.orbit/plugins/ to add custom analysis
    dimensions to orbit analyze.

    Example:
        # ~/.orbit/plugins/gripper_wear.py
        from orbit.plugins import BaseAnalyzer, AnalyzerResult

        class GripperWearAnalyzer(BaseAnalyzer):
            name = "gripper_wear"
            description = "Detect gripper wear patterns in action data"

            def analyze(self, action_episodes, info=None, **kwargs):
                # Your custom analysis logic
                score = 0.85
                return AnalyzerResult(
                    name=self.name,
                    score=score,
                    status="ok" if score > 0.7 else "warn",
                    summary="Gripper actions look healthy",
                    details={"wear_index": 0.15},
                )
    """

    name: str = "unnamed"
    description: str = ""

    @abstractmethod
    def analyze(
        self,
        action_episodes: list[np.ndarray],
        info: Any = None,
        **kwargs,
    ) -> "AnalyzerResult":
        """Run the analysis. Must return an AnalyzerResult."""
        ...


@dataclass
class AnalyzerResult:
    """Result from a custom analyzer plugin."""

    name: str
    score: float  # 0.0 - 1.0
    status: str  # "ok", "warn", "bad"
    summary: str
    details: dict | None = None


def discover_plugins(plugins_dir: Path | None = None) -> list[type[BaseAnalyzer]]:
    """Discover and load analyzer plugins from the plugins directory."""
    dir_path = plugins_dir or DEFAULT_PLUGINS_DIR
    if not dir_path.exists():
        return []

    analyzers: list[type[BaseAnalyzer]] = []

    for py_file in sorted(dir_path.glob("*.py")):
        if py_file.name.startswith("_"):
            continue

        try:
            spec = importlib.util.spec_from_file_location(
                f"orbit_plugin_{py_file.stem}", py_file
            )
            if spec is None or spec.loader is None:
                continue
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Find all BaseAnalyzer subclasses in the module
            for attr_name in dir(module):
                attr = getattr(module, attr_name)
                if (
                    isinstance(attr, type)
                    and issubclass(attr, BaseAnalyzer)
                    and attr is not BaseAnalyzer
                ):
                    analyzers.append(attr)
                    logger.debug(f"Loaded plugin: {attr.name} from {py_file.name}")

        except Exception as e:
            logger.warning(f"Failed to load plugin {py_file.name}: {e}")

    return analyzers


def run_plugins(
    action_episodes: list[np.ndarray],
    info: Any = None,
    plugins_dir: Path | None = None,
    **kwargs,
) -> list[AnalyzerResult]:
    """Discover and run all analyzer plugins, returning their results."""
    plugin_classes = discover_plugins(plugins_dir)
    results: list[AnalyzerResult] = []

    for cls in plugin_classes:
        try:
            analyzer = cls()
            result = analyzer.analyze(action_episodes, info=info, **kwargs)
            results.append(result)
        except Exception as e:
            logger.warning(f"Plugin {cls.name} failed: {e}")
            results.append(AnalyzerResult(
                name=cls.name,
                score=0.0,
                status="bad",
                summary=f"Plugin error: {e}",
            ))

    return results



"""LeRobot dataset loader for HuggingFace Hub."""

from __future__ import annotations

import json
import logging
import re

import numpy as np
import pandas as pd
from huggingface_hub import hf_hub_download
from huggingface_hub.hf_api import HfApi
from huggingface_hub.utils import (
    EntryNotFoundError,
    HfHubHTTPError,
    RepositoryNotFoundError,
)

from orbit.analyzer.dataset_loader import DatasetInfo
from orbit.loaders.base import BaseLoader, normalize_dataframe, normalize_shapes, sample_episode_indices

logger = logging.getLogger(__name__)


class LeRobotHubLoader(BaseLoader):
    """Load LeRobot datasets from HuggingFace Hub."""

    def can_load(self, source: str) -> bool:
        """True if source looks like a HuggingFace repo ID (owner/dataset)."""
        if not source or "/" not in source:
            return False
        # Must not be a local filesystem path
        if source.startswith("/") or source.startswith(".") or source.startswith("~"):
            return False
        # Must look like owner/name (exactly one slash, no path separators beyond that)
        parts = source.split("/")
        return len(parts) == 2 and all(p for p in parts)

    def load_info(self, source: str) -> DatasetInfo:
        """Fetch dataset metadata from HuggingFace Hub."""
        try:
            info_path = hf_hub_download(
                repo_id=source,
                filename="meta/info.json",
                repo_type="dataset",
            )
        except RepositoryNotFoundError:
            raise ValueError(f"Dataset '{source}' not found on HuggingFace Hub")
        except EntryNotFoundError:
            raise ValueError(f"'{source}' does not appear to be a LeRobot dataset")
        except (HfHubHTTPError, Exception) as e:
            if isinstance(e, ValueError):
                raise
            raise ConnectionError(
                f"Network error downloading metadata from '{source}': {e}"
            ) from e

        with open(info_path) as f:
            info = json.load(f)

        # Download episodes metadata (try v3 parquet first, then v1 jsonl)
        episodes_parquet_path = None
        episodes_path = None
        try:
            episodes_parquet_path = hf_hub_download(
                repo_id=source,
                filename="meta/episodes/chunk-000/file-000.parquet",
                repo_type="dataset",
            )
        except (EntryNotFoundError, Exception) as e:
            logger.debug("Episodes parquet not found, trying jsonl: %s", e)

        if episodes_parquet_path is None:
            try:
                episodes_path = hf_hub_download(
                    repo_id=source,
                    filename="meta/episodes.jsonl",
                    repo_type="dataset",
                )
            except EntryNotFoundError:
                episodes_path = None
            except (HfHubHTTPError, Exception) as e:
                if isinstance(e, (ValueError, EntryNotFoundError)):
                    raise
                raise ConnectionError(
                    f"Network error downloading episodes from '{source}': {e}"
                ) from e

        num_episodes = info.get("total_episodes", 0)
        num_frames = info.get("total_frames", 0)
        fps = float(info.get("fps", 0))
        robot_type = info.get("robot_type", "unknown")

        video_keys: list[str] = []
        camera_names: list[str] = []
        shapes: dict[str, list[int]] = {}

        features = info.get("features", {})
        for key, feat in features.items():
            if feat.get("dtype") == "video":
                video_keys.append(key)
                parts = key.split(".")
                camera_names.append(parts[-1])
            shape = feat.get("shape")
            if shape is not None:
                shapes[key] = shape

        # Override episode count from episodes metadata if available
        if episodes_parquet_path is not None:
            try:
                ep_df = pd.read_parquet(episodes_parquet_path)
                if len(ep_df) > 0:
                    num_episodes = len(ep_df)
            except Exception as e:
                logger.debug("Failed to read episodes parquet: %s", e)
        elif episodes_path is not None:
            from pathlib import Path

            ep_path = Path(episodes_path)
            if ep_path.exists():
                with open(ep_path) as f:
                    lines = [line.strip() for line in f if line.strip()]
                if lines:
                    num_episodes = len(lines)

        duration_seconds = num_frames / fps if fps > 0 else 0.0

        # Ensure canonical action/state keys exist in shapes
        normalize_shapes(shapes)

        return DatasetInfo(
            repo_id=source,
            num_episodes=num_episodes,
            num_frames=num_frames,
            fps=fps,
            robot_type=robot_type,
            camera_names=camera_names,
            video_keys=video_keys,
            shapes=shapes,
            duration_seconds=duration_seconds,
        )

    def load_episodes(
        self,
        source: str,
        episode_indices: list[int] | None = None,
        max_episodes: int | None = None,
    ) -> pd.DataFrame:
        """Download and load episode parquet files from a LeRobot dataset."""
        info = self.load_info(source)

        if max_episodes is None:
            max_episodes = 100

        # For large datasets, try direct filename construction (fast path)
        target_indices = episode_indices
        if target_indices is None and info.num_episodes > max_episodes:
            target_indices = sample_episode_indices(info.num_episodes, max_episodes)

        if target_indices is not None and info.num_episodes > 200:
            constructed = self._try_direct_episode_download(source, target_indices)
            if constructed:
                logger.info(
                    "Using direct episode download for %d/%d episodes",
                    len(target_indices),
                    info.num_episodes,
                )
                result = self._load_chunk_format(source, constructed, info, None, max_episodes)
                if not result.empty:
                    return normalize_dataframe(result)

        # Fallback: discover files
        data_files = self._discover_data_files(source, max_files=500)

        if not data_files:
            logger.warning("No data files found in %s", source)
            return normalize_dataframe(pd.DataFrame())

        is_chunk_format = any("chunk-" in f for f in data_files[:5])

        if is_chunk_format:
            result = self._load_chunk_format(
                source, data_files, info, episode_indices, max_episodes
            )
        else:
            result = self._load_per_episode_format(
                source, data_files, info, episode_indices, max_episodes
            )

        return normalize_dataframe(result)

    # --- Private helpers (extracted from dataset_loader.py) ---

    def _discover_data_files(
        self, repo_id: str, max_files: int = 500, max_iterations: int = 5000
    ) -> list[str]:
        api = HfApi()
        try:
            all_files = api.list_repo_files(repo_id, repo_type="dataset")
        except Exception:
            return []

        parquet_files = []
        iterations = 0
        for f in all_files:
            iterations += 1
            if iterations > max_iterations:
                break
            if f.startswith("data/") and f.endswith(".parquet"):
                parquet_files.append(f)
                if len(parquet_files) >= max_files:
                    break

        return sorted(parquet_files)

    def _construct_episode_filenames(
        self,
        repo_id: str,
        episode_indices: list[int],
        data_files_sample: list[str],
    ) -> list[str]:
        if not data_files_sample:
            return []

        sample = data_files_sample[0]

        # v3 per-episode: data/chunk-NNN/episode_NNNNNN.parquet
        m = re.match(r"data/chunk-(\d+)/episode_(\d+)\.parquet", sample)
        if m:
            ep_per_chunk = 1000
            chunk_nums = set()
            ep_nums = set()
            for f in data_files_sample[:50]:
                m2 = re.match(r"data/chunk-(\d+)/episode_(\d+)\.parquet", f)
                if m2:
                    chunk_nums.add(int(m2.group(1)))
                    ep_nums.add(int(m2.group(2)))
            if len(chunk_nums) > 1:
                sorted_eps = sorted(ep_nums)
                for i in range(1, len(sorted_eps)):
                    if sorted_eps[i] - sorted_eps[i - 1] > 1:
                        ep_per_chunk = sorted_eps[i]
                        break
            elif len(ep_nums) > 1:
                max_ep = max(ep_nums)
                if max_ep >= 1000:
                    ep_per_chunk = 1000

            filenames = []
            for ep_idx in episode_indices:
                chunk = ep_idx // ep_per_chunk
                filenames.append(f"data/chunk-{chunk:03d}/episode_{ep_idx:06d}.parquet")
            return filenames

        # v1: data/episode_NNNNNN.parquet
        m = re.match(r"data/episode_(\d+)\.parquet", sample)
        if m:
            return [f"data/episode_{ep_idx:06d}.parquet" for ep_idx in episode_indices]

        return []

    def _try_direct_episode_download(
        self, repo_id: str, episode_indices: list[int]
    ) -> list[str] | None:
        ep_per_chunk = 1000
        test_ep = episode_indices[0]
        test_chunk = test_ep // ep_per_chunk
        test_filename = f"data/chunk-{test_chunk:03d}/episode_{test_ep:06d}.parquet"

        try:
            hf_hub_download(
                repo_id=repo_id, filename=test_filename, repo_type="dataset"
            )
        except (EntryNotFoundError, Exception):
            test_filename = f"data/episode_{test_ep:06d}.parquet"
            try:
                hf_hub_download(
                    repo_id=repo_id, filename=test_filename, repo_type="dataset"
                )
                return [f"data/episode_{ep:06d}.parquet" for ep in episode_indices]
            except (EntryNotFoundError, Exception):
                return None

        return [
            f"data/chunk-{ep // ep_per_chunk:03d}/episode_{ep:06d}.parquet"
            for ep in episode_indices
        ]

    @staticmethod
    def _extract_episode_index_from_filename(filename: str) -> int | None:
        match = re.search(r"episode_(\d+)\.parquet$", filename)
        if match:
            return int(match.group(1))
        return None

    def _load_chunk_format(
        self,
        repo_id: str,
        data_files: list[str],
        info: DatasetInfo,
        episode_indices: list[int] | None,
        max_episodes: int,
    ) -> pd.DataFrame:
        is_per_episode = any("episode_" in f for f in data_files[:5])

        if is_per_episode:
            target_indices = episode_indices
            if target_indices is None and info.num_episodes > max_episodes:
                target_indices = sample_episode_indices(info.num_episodes, max_episodes)

            if target_indices is not None:
                target_set = set(target_indices)
                data_files = [
                    f
                    for f in data_files
                    if self._extract_episode_index_from_filename(f) in target_set
                ]

        if not is_per_episode and len(data_files) > 3:
            data_files = data_files[:3]

        frames: list[pd.DataFrame] = []
        for filename in data_files:
            try:
                parquet_path = hf_hub_download(
                    repo_id=repo_id, filename=filename, repo_type="dataset"
                )
            except EntryNotFoundError:
                continue
            except (HfHubHTTPError, Exception) as e:
                if isinstance(e, (ValueError, EntryNotFoundError)):
                    raise
                raise ConnectionError(
                    f"Network error downloading {filename} from '{repo_id}': {e}"
                ) from e

            df = pd.read_parquet(parquet_path)
            frames.append(df)

        if not frames:
            return pd.DataFrame()

        result = pd.concat(frames, ignore_index=True)

        if "episode_index" not in result.columns:
            return result

        if not is_per_episode:
            if episode_indices is not None:
                result = result[result["episode_index"].isin(episode_indices)]
            elif info.num_episodes > max_episodes:
                sampled = sample_episode_indices(info.num_episodes, max_episodes)
                result = result[result["episode_index"].isin(sampled)]

        return result

    def _load_per_episode_format(
        self,
        repo_id: str,
        data_files: list[str],
        info: DatasetInfo,
        episode_indices: list[int] | None,
        max_episodes: int,
    ) -> pd.DataFrame:
        if episode_indices is None:
            episode_indices = sample_episode_indices(info.num_episodes, max_episodes)

        frames: list[pd.DataFrame] = []
        for ep_idx in episode_indices:
            filename = f"data/episode_{ep_idx:06d}.parquet"
            if filename not in data_files:
                continue
            try:
                parquet_path = hf_hub_download(
                    repo_id=repo_id, filename=filename, repo_type="dataset"
                )
            except EntryNotFoundError:
                continue
            except (HfHubHTTPError, Exception) as e:
                if isinstance(e, (ValueError, EntryNotFoundError)):
                    raise
                raise ConnectionError(
                    f"Network error downloading episode {ep_idx} from '{repo_id}': {e}"
                ) from e

            df = pd.read_parquet(parquet_path)
            if "episode_index" not in df.columns:
                df["episode_index"] = ep_idx
            frames.append(df)

        if not frames:
            return pd.DataFrame()

        return pd.concat(frames, ignore_index=True)



"""Coverage analysis for LeRobot datasets."""

from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np
import pandas as pd

from orbit.analyzer.dataset_loader import DatasetInfo


@dataclass
class FullCoverageReport:
    """Coverage analysis results across multiple dimensions."""

    position_diversity: float  # 0.0 to 1.0
    action_diversity: float  # 0.0 to 1.0
    episode_consistency: float  # 0.0 to 1.0
    temporal_coverage: float  # 0.0 to 1.0

    # Detail fields for recommendations
    sparse_areas: list[str] = field(default_factory=list)
    low_variance_dims: list[str] = field(default_factory=list)
    outlier_episodes: list[int] = field(default_factory=list)
    uniform_action_pct: float = 0.0
    dominant_path_pct: float = 0.0


def compute_coverage(info: DatasetInfo, episode_data: pd.DataFrame) -> FullCoverageReport:
    """Compute coverage metrics from episode data.

    Args:
        info: Dataset metadata.
        episode_data: DataFrame with episode_index, action, state columns.

    Returns:
        FullCoverageReport with all coverage metrics.
    """
    if episode_data.empty:
        return FullCoverageReport(
            position_diversity=0.0,
            action_diversity=0.0,
            episode_consistency=0.0,
            temporal_coverage=0.0,
        )

    position_div, sparse_areas = _compute_position_diversity(episode_data)
    action_div, low_var_dims, uniform_pct = _compute_action_diversity(episode_data)
    consistency, outliers = _compute_episode_consistency(episode_data)
    temporal, dominant_pct = _compute_temporal_coverage(episode_data)

    return FullCoverageReport(
        position_diversity=position_div,
        action_diversity=action_div,
        episode_consistency=consistency,
        temporal_coverage=temporal,
        sparse_areas=sparse_areas,
        low_variance_dims=low_var_dims,
        outlier_episodes=outliers,
        uniform_action_pct=uniform_pct,
        dominant_path_pct=dominant_pct,
    )


def _extract_array(series: pd.Series) -> np.ndarray | None:
    """Extract numeric array from a column that may contain lists or arrays."""
    if series.empty:
        return None
    first = series.iloc[0]
    if first is None:
        return None
    if isinstance(first, (list, np.ndarray)):
        try:
            return np.array(series.tolist(), dtype=float)
        except (ValueError, TypeError):
            return None
    return None


def _compute_position_diversity(df: pd.DataFrame) -> tuple[float, list[str]]:
    """Compute how well the dataset covers the position space."""
    state_arr = _extract_array(df["state"]) if "state" in df.columns else None
    if state_arr is None or state_arr.size == 0:
        return 0.5, []

    # Use ALL state dimensions for diversity — there is no universal state
    # ordering (e.g. Austin BUDS: dims 0-2 are joint angles, not XYZ;
    # DLR Sara Pour: dims 0-2 ARE XYZ).  Using the full state vector
    # measures state-space coverage without assuming structure.
    if state_arr.ndim == 1:
        pos = state_arr.reshape(-1, 1)
    else:
        pos = state_arr

    # Discretize into grid and measure coverage
    n_bins = 5
    sparse_areas = []
    coverages = []

    for dim_idx in range(pos.shape[1]):
        col = pos[:, dim_idx]
        col_range = col.max() - col.min()
        if col_range < 1e-8:
            coverages.append(0.0)
            continue
        bins = np.linspace(col.min(), col.max(), n_bins + 1)
        counts, _ = np.histogram(col, bins=bins)
        filled = float(np.sum(counts > 0) / n_bins)
        coverages.append(filled)

        # Identify sparse bins
        threshold = len(col) / (n_bins * 3)
        for b_idx, count in enumerate(counts):
            if count < threshold:
                labels = ["left", "center-left", "center", "center-right", "right"]
                if b_idx < len(labels):
                    sparse_areas.append(f"the {labels[b_idx]} region (dimension {dim_idx})")

    diversity = float(np.mean(coverages)) if coverages else 0.5
    return min(1.0, diversity), sparse_areas


def _compute_action_diversity(df: pd.DataFrame) -> tuple[float, list[str], float]:
    """Compute how diverse the actions are across episodes."""
    action_arr = _extract_array(df["action"]) if "action" in df.columns else None
    if action_arr is None or action_arr.size == 0:
        return 0.5, [], 0.0

    n_dims = action_arr.shape[1] if action_arr.ndim == 2 else 1
    low_var_dims = []
    diversity_scores = []

    for dim_idx in range(n_dims):
        col = action_arr[:, dim_idx] if action_arr.ndim == 2 else action_arr
        std = np.std(col)
        col_range = np.ptp(col)
        # Coefficient of variation relative to range
        if col_range > 1e-8:
            cv = std / col_range
            diversity_scores.append(min(1.0, cv * 4))  # Scale so 0.25 cv -> 1.0
        else:
            diversity_scores.append(0.0)
            low_var_dims.append(f"dimension {dim_idx}")

    overall = float(np.mean(diversity_scores)) if diversity_scores else 0.5
    # Compute uniform action percentage
    uniform_count = sum(1 for s in diversity_scores if s < 0.2)
    uniform_pct = uniform_count / max(1, len(diversity_scores)) * 100

    return min(1.0, overall), low_var_dims, uniform_pct


def _compute_episode_consistency(df: pd.DataFrame) -> tuple[float, list[int]]:
    """Compute consistency across episodes and find outliers."""
    if "episode_index" not in df.columns:
        return 0.5, []

    episodes = df.groupby("episode_index")
    lengths = episodes.size()

    if len(lengths) < 2:
        return 1.0, []

    mean_len = lengths.mean()
    std_len = lengths.std()

    # Outliers are episodes whose length deviates > 2 std from the mean
    outliers = []
    if std_len > 0:
        z_scores = np.abs((lengths - mean_len) / std_len)
        outliers = [int(idx) for idx, z in z_scores.items() if z > 2.0]

    # Consistency = 1 - normalized coefficient of variation
    cv = std_len / mean_len if mean_len > 0 else 0
    consistency = max(0.0, min(1.0, 1.0 - cv))

    return consistency, outliers


def _compute_temporal_coverage(df: pd.DataFrame) -> tuple[float, float]:
    """Compute temporal/trajectory diversity using numpy (no scipy dependency)."""
    action_arr = _extract_array(df["action"]) if "action" in df.columns else None
    if action_arr is None or "episode_index" not in df.columns:
        return 0.5, 50.0

    episodes = df.groupby("episode_index")
    if len(episodes) < 2:
        return 0.5, 100.0

    # Compute per-episode mean action as a trajectory fingerprint
    fingerprints = []
    for _, ep_df in episodes:
        ep_actions = _extract_array(ep_df["action"])
        if ep_actions is not None and ep_actions.size > 0:
            fingerprints.append(np.mean(ep_actions, axis=0))

    if len(fingerprints) < 2:
        return 0.5, 100.0

    fingerprints_arr = np.array(fingerprints)
    n = len(fingerprints_arr)

    # Subsample for speed: cap at 100 fingerprints for pairwise computation
    if n > 100:
        rng = np.random.RandomState(42)
        idx = rng.choice(n, 100, replace=False)
        fingerprints_arr = fingerprints_arr[idx]
        n = 100

    # Pairwise distance using numpy broadcasting (avoids scipy import)
    # diff[i,j] = fingerprints[i] - fingerprints[j]
    diff = fingerprints_arr[:, np.newaxis, :] - fingerprints_arr[np.newaxis, :, :]
    dist_matrix = np.sqrt(np.sum(diff ** 2, axis=2))

    # Extract upper triangle (pairwise distances without duplicates/diagonal)
    triu_idx = np.triu_indices(n, k=1)
    distances = dist_matrix[triu_idx]

    if len(distances) == 0:
        return 0.5, 100.0

    # Diversity = normalized spread of distances
    mean_dist = np.mean(distances)
    max_possible = np.max(distances) if np.max(distances) > 0 else 1.0
    temporal_score = min(1.0, mean_dist / max_possible) if max_possible > 0 else 0.0

    # Dominant path: count how many episodes are within small distance of mode
    median_dist = np.median(distances)
    close_threshold = median_dist * 0.5
    close_counts = np.sum(dist_matrix < close_threshold, axis=1)
    dominant_pct = float(np.max(close_counts) / n * 100)

    return temporal_score, dominant_pct



"""Structured Gemini Vision quality assessment for robot demonstration datasets.

Sends sampled frames from episodes to Gemini Flash with specific quality questions.
Only called when the dataset has image observations. Skips gracefully for state-only datasets.

Cost: ~$0.003 per analysis (9 images to Gemini Flash).
"""

from __future__ import annotations

import json
import logging
import os
from dataclasses import dataclass, field
from typing import Any

import numpy as np

logger = logging.getLogger(__name__)


QUALITY_ASSESSMENT_PROMPT = """You are analyzing camera frames from a robot manipulation dataset.
I will show you frames from 3 different episodes at 3 timepoints each (start, middle, end).

For each episode, answer these questions with YES/NO and a confidence score (0.0-1.0):

1. TASK_VISIBLE: Can you see the robot performing a clear manipulation task?
2. TASK_COMPLETE: Does the final frame show what appears to be a completed task (object moved, assembled, etc)?
3. VISUAL_QUALITY: Are the camera images clear (not blurry, not occluded, good lighting)?
4. ANOMALY: Does any episode look clearly different or problematic compared to the others?

Also answer this cross-episode question:
5. CONSISTENT_TASK: Do all 3 episodes appear to be attempting the SAME task?

Respond ONLY in this JSON format:
{
  "episodes": [
    {
      "episode_index": 0,
      "task_visible": {"answer": true, "confidence": 0.9},
      "task_complete": {"answer": true, "confidence": 0.7},
      "visual_quality": {"answer": true, "confidence": 0.95},
      "anomaly": {"answer": false, "confidence": 0.85}
    }
  ],
  "consistent_task": {"answer": true, "confidence": 0.9},
  "task_description": "Robot arm picking up a red cube and placing it in a bin",
  "concerns": ["Episode 2 appears to end before task completion"]
}
"""


@dataclass
class VisionQualityResult:
    """Result of structured vision quality assessment."""

    task_visible_all: bool  # All episodes show visible task
    task_complete_fraction: float  # Fraction of episodes with completed task
    consistent_task: bool  # Episodes show same task
    visual_quality_all: bool  # All episodes have good visual quality
    anomaly_episodes: list[int]  # Episodes flagged as anomalous
    task_description: str  # VLM's description of the task
    concerns: list[str]  # Specific concerns raised
    grade_adjustments: dict[str, int]  # penalty name -> points to subtract
    total_penalty: int  # Sum of all penalties
    raw_response: dict = field(default_factory=dict)
    model_used: str = "gemini"
    error: str | None = None  # Set if assessment failed


def _has_image_observations(dataset_info: Any) -> bool:
    """Check if dataset has image observations."""
    if hasattr(dataset_info, "observation_keys"):
        for key in (dataset_info.observation_keys or []):
            if "image" in key.lower():
                return True
    if hasattr(dataset_info, "camera_names"):
        if dataset_info.camera_names:
            return True
    if hasattr(dataset_info, "features"):
        features = dataset_info.features or {}
        for key in features:
            if "image" in key.lower():
                return True
    return False


def _select_episodes(
    episode_qualities: dict[int, float] | None,
    total_episodes: int,
) -> list[int]:
    """Select 3 episodes: best quality, worst quality, and random middle.

    Args:
        episode_qualities: Mapping of episode index to quality score.
            If None, picks indices 0, total//2, total-1.
        total_episodes: Total number of episodes.

    Returns:
        List of 3 episode indices.
    """
    if total_episodes <= 0:
        return []
    if total_episodes <= 3:
        return list(range(total_episodes))

    if episode_qualities and len(episode_qualities) >= 3:
        sorted_eps = sorted(episode_qualities.items(), key=lambda x: x[1])
        worst = sorted_eps[0][0]
        best = sorted_eps[-1][0]
        # Pick a middle episode (not best or worst)
        middle_candidates = [idx for idx, _ in sorted_eps[1:-1]]
        if middle_candidates:
            rng = np.random.RandomState(42)
            middle = middle_candidates[rng.randint(len(middle_candidates))]
        else:
            middle = sorted_eps[len(sorted_eps) // 2][0]
        return [best, middle, worst]

    # Fallback: evenly spaced
    return [0, total_episodes // 2, total_episodes - 1]


def _extract_frame_indices(episode_length: int) -> list[int]:
    """Extract frame indices at 10%, 50%, 90% of episode length."""
    if episode_length <= 0:
        return []
    if episode_length <= 3:
        return list(range(episode_length))
    return [
        int(episode_length * 0.10),
        int(episode_length * 0.50),
        int(episode_length * 0.90),
    ]


def _parse_response(response_text: str) -> dict:
    """Parse JSON response from Gemini, handling markdown code blocks."""
    text = response_text.strip()
    # Strip markdown code block if present
    if text.startswith("```"):
        lines = text.split("\n")
        # Remove first and last lines (```json and ```)
        lines = [l for l in lines if not l.strip().startswith("```")]
        text = "\n".join(lines)
    try:
        return json.loads(text)
    except json.JSONDecodeError:
        logger.warning("Failed to parse Gemini vision quality response as JSON")
        return {}


def _compute_adjustments(parsed: dict) -> dict[str, int]:
    """Map parsed VLM response to grade adjustments."""
    adjustments: dict[str, int] = {}

    episodes = parsed.get("episodes", [])
    if not episodes:
        return adjustments

    # task_visible: if any episode has task_visible=False, penalize
    invisible = [e for e in episodes if not e.get("task_visible", {}).get("answer", True)]
    if invisible:
        adjustments["camera_misconfigured"] = -5

    # task_complete: if >50% episodes incomplete
    incomplete = [e for e in episodes if not e.get("task_complete", {}).get("answer", True)]
    if len(incomplete) > len(episodes) / 2:
        adjustments["incomplete_tasks"] = -10

    # consistent_task
    consistent = parsed.get("consistent_task", {})
    if not consistent.get("answer", True):
        adjustments["inconsistent_tasks"] = -8

    # visual_quality
    poor_quality = [e for e in episodes if not e.get("visual_quality", {}).get("answer", True)]
    if poor_quality:
        adjustments["poor_visual_quality"] = -3

    return adjustments


def assess_vision_quality(
    dataset_info: Any,
    frame_loader: Any | None = None,
    episode_qualities: dict[int, float] | None = None,
) -> VisionQualityResult:
    """Run structured Gemini vision quality assessment.

    Args:
        dataset_info: DatasetInfo with metadata about the dataset.
        frame_loader: Callable(episode_idx, frame_idx) -> PIL.Image or np.ndarray.
            If None, assessment is skipped.
        episode_qualities: Optional per-episode quality scores for episode selection.

    Returns:
        VisionQualityResult. If no images or no API key, returns a no-op result with zero penalty.
    """
    # Check if dataset has images
    if not _has_image_observations(dataset_info):
        return VisionQualityResult(
            task_visible_all=True,
            task_complete_fraction=1.0,
            consistent_task=True,
            visual_quality_all=True,
            anomaly_episodes=[],
            task_description="State-only dataset — no visual assessment",
            concerns=[],
            grade_adjustments={},
            total_penalty=0,
            error="No image observations in dataset",
        )

    if frame_loader is None:
        return VisionQualityResult(
            task_visible_all=True,
            task_complete_fraction=1.0,
            consistent_task=True,
            visual_quality_all=True,
            anomaly_episodes=[],
            task_description="Frame loader not available",
            concerns=[],
            grade_adjustments={},
            total_penalty=0,
            error="No frame loader provided",
        )

    api_key = os.environ.get("GOOGLE_API_KEY")
    if not api_key:
        return VisionQualityResult(
            task_visible_all=True,
            task_complete_fraction=1.0,
            consistent_task=True,
            visual_quality_all=True,
            anomaly_episodes=[],
            task_description="No API key available",
            concerns=[],
            grade_adjustments={},
            total_penalty=0,
            error="GOOGLE_API_KEY not set",
        )

    num_episodes = getattr(dataset_info, "num_episodes", 0)
    selected = _select_episodes(episode_qualities, num_episodes)
    if not selected:
        return VisionQualityResult(
            task_visible_all=True,
            task_complete_fraction=1.0,
            consistent_task=True,
            visual_quality_all=True,
            anomaly_episodes=[],
            task_description="No episodes to assess",
            concerns=[],
            grade_adjustments={},
            total_penalty=0,
            error="No episodes available",
        )

    # Load frames
    import base64
    import io

    frames_data = []
    for ep_idx in selected:
        ep_length = getattr(dataset_info, "avg_episode_length", 100)
        frame_indices = _extract_frame_indices(int(ep_length))
        for f_idx in frame_indices:
            try:
                frame = frame_loader(ep_idx, f_idx)
                if frame is not None:
                    # Convert to base64 for API
                    if hasattr(frame, "save"):
                        # PIL Image
                        buf = io.BytesIO()
                        frame.save(buf, format="JPEG")
                        frames_data.append(base64.b64encode(buf.getvalue()).decode())
                    elif isinstance(frame, np.ndarray):
                        from PIL import Image
                        img = Image.fromarray(frame.astype(np.uint8))
                        buf = io.BytesIO()
                        img.save(buf, format="JPEG")
                        frames_data.append(base64.b64encode(buf.getvalue()).decode())
            except Exception as e:
                logger.debug(f"Failed to load frame ep={ep_idx} f={f_idx}: {e}")

    if not frames_data:
        return VisionQualityResult(
            task_visible_all=True,
            task_complete_fraction=1.0,
            consistent_task=True,
            visual_quality_all=True,
            anomaly_episodes=[],
            task_description="Could not load any frames",
            concerns=[],
            grade_adjustments={},
            total_penalty=0,
            error="Frame loading failed",
        )

    # Call Gemini Flash
    try:
        import google.generativeai as genai

        genai.configure(api_key=api_key)
        model = genai.GenerativeModel("gemini-2.0-flash")

        # Build content with images
        content_parts = [QUALITY_ASSESSMENT_PROMPT]
        for i, b64 in enumerate(frames_data):
            ep_num = i // 3
            frame_num = i % 3
            frame_label = ["start", "middle", "end"][frame_num] if frame_num < 3 else f"frame_{frame_num}"
            content_parts.append(f"\nEpisode {ep_num + 1}, {frame_label} frame:")
            content_parts.append({
                "mime_type": "image/jpeg",
                "data": b64,
            })

        response = model.generate_content(content_parts)
        parsed = _parse_response(response.text)

    except Exception as e:
        logger.warning(f"Gemini vision quality assessment failed: {e}")
        return VisionQualityResult(
            task_visible_all=True,
            task_complete_fraction=1.0,
            consistent_task=True,
            visual_quality_all=True,
            anomaly_episodes=[],
            task_description="Assessment failed",
            concerns=[],
            grade_adjustments={},
            total_penalty=0,
            error=str(e),
        )

    # Parse results
    episodes = parsed.get("episodes", [])
    adjustments = _compute_adjustments(parsed)
    total_penalty = sum(adjustments.values())

    # Extract per-field summaries
    task_visible_all = all(
        e.get("task_visible", {}).get("answer", True) for e in episodes
    )
    complete_count = sum(
        1 for e in episodes if e.get("task_complete", {}).get("answer", True)
    )
    task_complete_fraction = complete_count / max(len(episodes), 1)
    consistent = parsed.get("consistent_task", {}).get("answer", True)
    visual_quality_all = all(
        e.get("visual_quality", {}).get("answer", True) for e in episodes
    )
    anomaly_eps = [
        selected[i] for i, e in enumerate(episodes)
        if e.get("anomaly", {}).get("answer", False) and i < len(selected)
    ]

    return VisionQualityResult(
        task_visible_all=task_visible_all,
        task_complete_fraction=task_complete_fraction,
        consistent_task=consistent,
        visual_quality_all=visual_quality_all,
        anomaly_episodes=anomaly_eps,
        task_description=parsed.get("task_description", "Unknown"),
        concerns=parsed.get("concerns", []),
        grade_adjustments=adjustments,
        total_penalty=total_penalty,
        raw_response=parsed,
        model_used="gemini-2.0-flash",
    )



"""Tests for orbit badge."""

from orbit.badge import generate_badge_markdown


def test_badge_markdown_basic():
    md = generate_badge_markdown("A")
    assert "ORBIT" in md
    assert "Grade%20A" in md
    assert "brightgreen" in md


def test_badge_markdown_with_score():
    md = generate_badge_markdown("B", score=72)
    assert "72/100" in md
    assert "green" in md


def test_badge_markdown_grade_f():
    md = generate_badge_markdown("F")
    assert "red" in md



"""Universal dataset loader registry with auto-detection.

Usage:
    from orbit.loaders import detect_and_load_info, detect_and_load_episodes

    info = detect_and_load_info("lerobot/pusht")                # HuggingFace
    info = detect_and_load_info("./my-dataset/")                # Local LeRobot
    info = detect_and_load_info("./demo.hdf5")                  # RoboMimic HDF5
    info = detect_and_load_info("./recording.mcap")             # ROS 2 MCAP
    info = detect_and_load_info("./data/")                      # Raw numpy/csv/parquet

    # Universal loader with intelligent array profiling:
    from orbit.loaders.universal import universal_load
    info, df, profiles = universal_load("lerobot/pusht")
"""

from __future__ import annotations

import logging
from pathlib import Path

import pandas as pd

from orbit.analyzer.dataset_loader import DatasetInfo
from orbit.loaders.base import BaseLoader

logger = logging.getLogger(__name__)

# Lazy-loaded loader instances (created on first use)
_loader_cache: dict[str, BaseLoader] = {}


def _get_loaders() -> list[tuple[str, BaseLoader]]:
    """Return all registered loaders in detection priority order."""
    from orbit.loaders.directory import DirectoryLoader
    from orbit.loaders.hdf5 import HDF5Loader
    from orbit.loaders.lerobot_hub import LeRobotHubLoader
    from orbit.loaders.lerobot_local import LeRobotLocalLoader
    from orbit.loaders.rlds import RLDSLoader
    from orbit.loaders.rosbag import RosBagLoader

    # Order: most specific first
    return [
        ("lerobot_local", LeRobotLocalLoader()),
        ("hdf5", HDF5Loader()),
        ("rosbag", RosBagLoader()),
        ("rlds", RLDSLoader()),
        ("directory", DirectoryLoader()),
        ("lerobot_hub", LeRobotHubLoader()),
    ]


# Map of format names to loader classes for explicit --format override
FORMAT_NAMES: dict[str, type] = {}


def _init_format_names():
    global FORMAT_NAMES
    if FORMAT_NAMES:
        return
    from orbit.loaders.directory import DirectoryLoader
    from orbit.loaders.hdf5 import HDF5Loader
    from orbit.loaders.lerobot_hub import LeRobotHubLoader
    from orbit.loaders.lerobot_local import LeRobotLocalLoader
    from orbit.loaders.rlds import RLDSLoader
    from orbit.loaders.rosbag import RosBagLoader

    FORMAT_NAMES = {
        "lerobot": LeRobotHubLoader,
        "lerobot-hub": LeRobotHubLoader,
        "lerobot-local": LeRobotLocalLoader,
        "rlds": RLDSLoader,
        "hdf5": HDF5Loader,
        "rosbag": RosBagLoader,
        "directory": DirectoryLoader,
    }


def detect_loader(
    source: str,
    format: str | None = None,
    **loader_kwargs,
) -> BaseLoader:
    """Detect the appropriate loader for the given source.

    Args:
        source: Path, file, or HuggingFace repo ID.
        format: Explicit format override (bypasses auto-detection).
        **loader_kwargs: Extra kwargs passed to loader constructor
            (e.g., action_topic, state_topic for RosBagLoader).

    Returns:
        A BaseLoader instance that can handle the source.

    Raises:
        ValueError: If no loader can handle the source.
    """
    # Explicit format override
    if format is not None:
        _init_format_names()
        fmt = format.lower().strip()
        if fmt == "auto":
            format = None  # Fall through to auto-detection
        elif fmt in FORMAT_NAMES:
            return FORMAT_NAMES[fmt](**loader_kwargs)
        else:
            raise ValueError(
                f"Unknown format '{format}'. "
                f"Supported formats: {', '.join(sorted(FORMAT_NAMES.keys()))}"
            )

    # Auto-detection: try each loader in priority order
    for name, loader in _get_loaders():
        try:
            if loader.can_load(source):
                # If loader needs kwargs (e.g. RosBagLoader), reinstantiate
                if loader_kwargs and hasattr(loader.__class__.__init__, '__code__'):
                    try:
                        return loader.__class__(**loader_kwargs)
                    except TypeError:
                        pass
                return loader
        except Exception:
            # If a loader's can_load raises (e.g. missing dependency), skip it
            continue

    # Provide helpful error message
    p = Path(source).expanduser()
    if p.exists():
        if p.is_dir():
            contents = list(p.iterdir())[:10]
            file_types = set(f.suffix for f in contents if f.is_file())
            raise ValueError(
                f"Could not detect dataset format in '{source}'.\n"
                f"  Directory contains: {file_types or 'no files'}\n"
                f"  Supported formats: LeRobot (meta/info.json), RLDS (.tfrecord), "
                f"numpy (.npz/.npy), CSV, parquet\n"
                f"  Tip: Use --format to specify the format explicitly."
            )
        else:
            raise ValueError(
                f"Could not detect dataset format for '{source}' "
                f"(extension: {p.suffix}).\n"
                f"  Supported file types: .hdf5, .h5, .bag, .mcap\n"
                f"  Tip: Use --format to specify the format explicitly."
            )
    else:
        # Not a local path — maybe a HuggingFace repo ID?
        if "/" in source:
            raise ValueError(
                f"Dataset '{source}' not found locally and could not be loaded "
                f"from HuggingFace Hub.\n"
                f"  Check the repo ID or provide a valid local path."
            )
        raise ValueError(
            f"'{source}' is not a valid path or HuggingFace repo ID.\n"
            f"  Examples:\n"
            f"    orbit analyze lerobot/pusht          (HuggingFace)\n"
            f"    orbit analyze ./my-dataset/           (local LeRobot)\n"
            f"    orbit analyze ./demo.hdf5             (RoboMimic HDF5)\n"
            f"    orbit analyze ./recording.mcap        (ROS 2 bag)"
        )


def detect_and_load_info(
    source: str,
    format: str | None = None,
    **loader_kwargs,
) -> DatasetInfo:
    """Auto-detect format and load dataset metadata.

    Args:
        source: Path, file, or HuggingFace repo ID.
        format: Explicit format override (bypasses auto-detection).
        **loader_kwargs: Extra kwargs passed to loader constructor.

    Returns:
        DatasetInfo with parsed metadata.
    """
    loader = detect_loader(source, format=format, **loader_kwargs)
    return loader.load_info(source)


def detect_and_load_episodes(
    source: str,
    episode_indices: list[int] | None = None,
    max_episodes: int | None = None,
    format: str | None = None,
    **loader_kwargs,
) -> pd.DataFrame:
    """Auto-detect format and load episode data.

    Args:
        source: Path, file, or HuggingFace repo ID.
        episode_indices: Specific episodes to load.
        max_episodes: Maximum episodes to sample.
        format: Explicit format override.
        **loader_kwargs: Extra kwargs passed to loader constructor.
            action_key / state_key are extracted and applied as column
            renames after loading.

    Returns:
        DataFrame with columns: [episode_index, frame_index, action, state, timestamp]
    """
    # Extract column override keys before passing to loader constructor
    action_key = loader_kwargs.pop("action_key", None)
    state_key = loader_kwargs.pop("state_key", None)

    loader = detect_loader(source, format=format, **loader_kwargs)
    df = loader.load_episodes(source, episode_indices=episode_indices, max_episodes=max_episodes)

    # Apply user-specified column renames if the canonical names are missing
    if action_key and "action" not in df.columns and action_key in df.columns:
        df = df.rename(columns={action_key: "action"})
    if state_key and "state" not in df.columns and state_key in df.columns:
        df = df.rename(columns={state_key: "state"})

    return df



"""Tests for orbit explore."""

from orbit.explorer import DatasetResult, print_explore_results


def test_dataset_result_creation():
    r = DatasetResult(
        repo_id="lerobot/test",
        robot_type="so100",
        num_episodes=50,
        task="pick and place",
        downloads=100,
        likes=5,
        grade="B",
    )
    assert r.repo_id == "lerobot/test"
    assert r.grade == "B"


def test_print_explore_results_empty(capsys):
    """Empty results should not crash."""
    print_explore_results([])



"""Tests for the multi-framework training configuration generator."""

from unittest.mock import MagicMock

import pytest

from orbit.training.base import FRAMEWORKS, TrainingConfig, get_framework_config


def _make_info(n_episodes=50, fps=30.0, action_dim=6):
    info = MagicMock()
    info.repo_id = "test/dataset"
    info.num_episodes = n_episodes
    info.num_frames = n_episodes * 100
    info.fps = fps
    info.robot_type = "so100"
    info.camera_names = ["observation.image"]
    info.shapes = {"action": [action_dim], "observation.image": [3, 480, 640]}
    return info


class TestFrameworkRegistry:
    def test_all_frameworks_listed(self):
        assert "lerobot" in FRAMEWORKS
        assert "openvla" in FRAMEWORKS
        assert "openpi" in FRAMEWORKS
        assert "groot" in FRAMEWORKS
        assert "osmo" in FRAMEWORKS
        assert "custom" in FRAMEWORKS

    def test_unknown_framework_raises(self):
        with pytest.raises(ValueError, match="Unknown framework"):
            get_framework_config(
                framework="nonexistent",
                dataset_repo="test/data",
                policy_type="act",
                info=_make_info(),
                episode_count=50,
                action_dim=6,
                camera_count=1,
                resolution="480p",
            )


class TestLeRobotConfig:
    def test_generates_lerobot_train_command(self):
        config = get_framework_config(
            framework="lerobot",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
            batch_size=32,
            steps=100_000,
            learning_rate=1e-4,
            chunk_size=50,
        )
        assert isinstance(config, TrainingConfig)
        assert config.framework == "lerobot"
        assert "lerobot-train" in config.command
        assert "--dataset.repo_id=test/dataset" in config.command
        assert "--policy.type=act" in config.command
        assert config.dataset_format == "lerobot-v3"
        assert not config.conversion_needed

    def test_pre_checks_include_gate(self):
        config = get_framework_config(
            framework="lerobot",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
        )
        assert any("orbit gate" in check for check in config.pre_checks)


class TestOpenVLAConfig:
    def test_generates_torchrun_command(self):
        config = get_framework_config(
            framework="openvla",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
            gpu_memory_gb=80,
        )
        assert config.framework == "openvla"
        assert "torchrun" in config.command
        assert "finetune.py" in config.command
        assert config.dataset_format == "rlds"
        assert config.conversion_needed

    def test_oft_recommended_for_40gb_plus(self):
        """OFT (Open Fine-Tuning) should be default for 40GB+ VRAM."""
        config = get_framework_config(
            framework="openvla",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
            gpu_memory_gb=80,
        )
        assert "openvla-7b-oft" in config.command
        assert any("OFT" in tip for tip in config.tips)

    def test_lora_fallback_for_low_vram(self):
        """LoRA should be used for GPUs with less than 40GB."""
        config = get_framework_config(
            framework="openvla",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
            gpu_memory_gb=24,
        )
        assert "openvla-7b-oft" not in config.command
        assert "lora_rank" in config.command


class TestOpenPIConfig:
    def test_generates_uv_run_command(self):
        config = get_framework_config(
            framework="openpi",
            dataset_repo="test/my-dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
        )
        assert config.framework == "openpi"
        assert "uv run scripts/train.py" in config.command
        assert config.dataset_format == "lerobot-v3"

    def test_generates_train_config_file(self):
        """OpenPI should generate a TrainConfig Python file template."""
        config = get_framework_config(
            framework="openpi",
            dataset_repo="test/my-dataset",
            policy_type="act",
            info=_make_info(action_dim=6),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
        )
        assert config.generated_files
        config_files = list(config.generated_files.keys())
        assert any("config" in f and f.endswith(".py") for f in config_files)

        # Check config content
        config_content = list(config.generated_files.values())[0]
        assert "TrainConfig" in config_content
        assert "action_dim" in config_content
        assert "6" in config_content  # action_dim value

    def test_config_includes_camera_keys(self):
        """Config should include actual camera keys from dataset."""
        info = _make_info()
        info.camera_names = ["cam_left", "cam_right"]
        config = get_framework_config(
            framework="openpi",
            dataset_repo="test/data",
            policy_type="act",
            info=info,
            episode_count=50,
            action_dim=6,
            camera_count=2,
            resolution="480p",
        )
        config_content = list(config.generated_files.values())[0]
        assert "cam_left" in config_content
        assert "cam_right" in config_content


class TestGR00TConfig:
    def test_uses_launch_finetune_command(self):
        """GR00T N1.6 must use gr00t/experiment/launch_finetune.py."""
        config = get_framework_config(
            framework="groot",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
        )
        assert config.framework == "groot"
        assert "launch_finetune.py" in config.command
        assert "GR00T-N1.6-3B" in config.command
        assert "--embodiment-tag" in config.command
        assert "--modality-config-path" in config.command

    def test_requires_lerobot_v2(self):
        """GR00T N1.6 requires LeRobot v2 format, NOT v3."""
        config = get_framework_config(
            framework="groot",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
        )
        assert config.dataset_format == "lerobot-v2"
        assert config.conversion_needed
        assert "lerobot-v2" in config.conversion_command

    def test_generates_modality_config(self):
        """Must generate modality config Python file."""
        config = get_framework_config(
            framework="groot",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
        )
        assert config.generated_files
        files = config.generated_files

        # Should have modality config Python file
        py_files = [f for f in files if f.endswith("_config.py")]
        assert len(py_files) >= 1
        py_content = files[py_files[0]]
        assert "ModalityConfig" in py_content

        # Should have modality.json
        json_files = [f for f in files if f.endswith("modality.json")]
        assert len(json_files) >= 1


class TestOSMOConfig:
    def test_generates_yaml_workflow(self):
        config = get_framework_config(
            framework="osmo",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
            batch_size=32,
            steps=100_000,
        )
        assert config.framework == "osmo"
        assert "workflow:" in config.command
        assert "orbit-gate" in config.command
        assert "train-policy" in config.command
        assert "orbit-verify" in config.command

    def test_osmo_includes_all_pipeline_stages(self):
        config = get_framework_config(
            framework="osmo",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
        )
        assert "orbit-analyze" in config.command
        assert "orbit-gate" in config.command
        assert "train-policy" in config.command
        assert "orbit-verify" in config.command


class TestCustomConfig:
    def test_generates_generic_command(self):
        config = get_framework_config(
            framework="custom",
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
            batch_size=8,
            steps=100_000,
            learning_rate=1e-4,
            fps=30.0,
        )
        assert config.framework == "custom"
        assert "python train.py" in config.command
        assert "--dataset test/dataset" in config.command


class TestAllFrameworksGeneration:
    @pytest.mark.parametrize("framework", FRAMEWORKS)
    def test_each_framework_returns_valid_config(self, framework):
        config = get_framework_config(
            framework=framework,
            dataset_repo="test/dataset",
            policy_type="act",
            info=_make_info(),
            episode_count=50,
            action_dim=6,
            camera_count=1,
            resolution="480p",
            gpu_memory_gb=80,
            batch_size=32,
            steps=100_000,
            learning_rate=1e-4,
            chunk_size=50,
            fps=30.0,
        )
        assert isinstance(config, TrainingConfig)
        assert config.framework == framework
        assert len(config.command) > 10
        assert config.estimated_time
        assert config.gpu_requirements
        assert config.dataset_format
